/** @file emu.h

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@note
	Modified for BML3MK5 by Sasaji at 2011.06.17

	@brief [ win32 emulation i/f ]

*/

#ifndef _EMU_H_
#define _EMU_H_

#include <windows.h>

#include "common.h"
#include "logging.h"
#include "utility.h"
#include "config.h"
#include "vm/vm.h"
#include "res/resource.h"
#include "clocale.h"
#include "cmutex.h"

// DirectX
#define DIRECTSOUND_VERSION 0x900
#ifdef USE_DIRECT3D
//#define DIRECT3D_VERSION 0x900
#include <d3d9.h>
#pragma comment(lib, "d3d9.lib")
#ifdef USE_DIRECT3DX
#include <d3dx9.h>
#pragma comment(lib, "d3dx9.lib")
#endif
#endif

#ifdef USE_DIRECTINPUT
#define DIRECTINPUT_VERSION	0x800
#include <dinput.h>
#if DIRECTINPUT_VERSION >= 0x0800
#pragma comment(lib, "dinput8.lib")
#else
#pragma comment(lib, "dinput.lib")
#endif
#endif

#define WM_RESIZE  (WM_USER + 1)

#ifndef SCREEN_WIDTH_ASPECT
#define SCREEN_WIDTH_ASPECT SCREEN_WIDTH
#endif
#ifndef SCREEN_HEIGHT_ASPECT
#define SCREEN_HEIGHT_ASPECT SCREEN_HEIGHT
#endif
#ifndef WINDOW_WIDTH
#define WINDOW_WIDTH SCREEN_WIDTH_ASPECT
#endif
#ifndef WINDOW_HEIGHT
#define WINDOW_HEIGHT SCREEN_HEIGHT_ASPECT
#endif

#define SCREEN_MODE_MAX (ID_SCREEN_FULLSCREEN24 - ID_SCREEN_FULLSCREEN1 + 1)
typedef struct screen_mode_st {
	int disp_no;
	int width;
	int height;
} screen_mode_t;
#define WINDOW_MODE_MAX (ID_SCREEN_WINDOW8 - ID_SCREEN_WINDOW1 + 1)
typedef struct window_mode_st {
	int power;
	int width;
	int height;
} window_mode_t;

#include <dsound.h>
#pragma comment(lib, "dxguid.lib")

#ifdef USE_SOCKET
#include <winsock.h>
#endif

#ifdef USE_MEDIA
#define MEDIA_MAX 64
#endif

#ifdef USE_SOCKET
#define SOCKET_MAX 6
#define SOCKET_BUFFER_MAX 0x100000
#define WM_SOCKET0 (WM_USER + 2)
#define WM_SOCKET1 (WM_USER + 3)
#define WM_SOCKET2 (WM_USER + 4)
#define WM_SOCKET3 (WM_USER + 5)
#define WM_SOCKET4 (WM_USER + 6)
#define WM_SOCKET5 (WM_USER + 7)
#endif

#define KEY_STATUS_SIZE 512
#define KEY_MOD_ALT_KEY 0x0001

#ifdef USE_FD1
// for d88 bank switch
#define MAX_D88_BANKS 50

typedef struct {
	_TCHAR name[20];
	int offset;
} d88_bank_t;

typedef struct {
	_TCHAR path[_MAX_PATH];
	d88_bank_t bank[MAX_D88_BANKS];
	int bank_nums;
	int cur_bank;
	int prev_bank;
} d88_file_t;
#endif

#ifdef USE_DEBUGGER

typedef struct {
	EMU *emu;
	VM *vm;
	int cpu_index;
	bool running;
	bool request_terminate;
} debugger_thread_t;

#define MAX_BREAK_POINTS	8

typedef	struct {
	uint32_t addr, mask;
	int len;	// range 0 - addr + len
	int status;	// 0 = none, 1 = enabled, other = disabled
} break_point_table_t;

typedef struct {
	break_point_table_t table[MAX_BREAK_POINTS];
	break_point_table_t stored[MAX_BREAK_POINTS];
	bool hit;
	uint32_t hit_addr;
} break_point_t;

#include "win32_debugger.h"

#endif

class FIFO;
class FILEIO;
class GUI;
//class LedBox;
class MsgBoard;
class CSurface;
class REC_VIDEO;
class REC_AUDIO;

/**
	@brief emulation i/f
*/
class EMU : public Logging
{
protected:
	VM* vm;
	GUI *gui;
	bool initialized;
private:
	// ----------------------------------------
	// input
	// ----------------------------------------
	/// @name input private methods
	//@{
	void EMU_INPUT();
	void initialize_input();
	void release_input();
	void update_input();

	void initialize_joystick();
	void release_joystick();

	void initialize_mouse(bool enable);
#ifdef USE_MOUSE_ABSOLUTE
	void set_mouse_position();
#endif

	int  translate_global_key(int code);
	//@}
	/// @name input private members
	//@{
#ifdef USE_KEYCODE_CONV
	uint8_t keycode_conv[KEY_STATUS_SIZE];
#endif
	uint8_t key_status[KEY_STATUS_SIZE];	// windows key code mapping
	/// vm key scan code mapping (provided from vm)
	/// bit0: set/reset on auto key
	/// bit4: set/reset on virtual key
	uint8_t *vm_key_status;
	int   vm_key_status_size;
#ifdef USE_SHIFT_NUMPAD_KEY
	uint8_t key_converted[KEY_STATUS_SIZE];
	bool key_shift_pressed, key_shift_released;
#endif
	bool lost_focus;
	int  key_mod;

#ifdef USE_JOYSTICK
	uint32_t joy_status[2];	// joystick #1, #2 (b0 = up, b1 = down, b2 = left, b3 = right, b4-b27 = trigger #1-#24
	int joy_num;
	int joyid_map[2];
	uint32_t joy_xmin[2], joy_xmax[2];
	uint32_t joy_ymin[2], joy_ymax[2];
	bool joy_enabled[2];
	bool use_joystick;
	INPUT dummy_key;
	int send_dummy_key;
#endif

	int mouse_status[3];	// x, y, button (b0 = left, b1 = right)
	int mouse_disabled;
#ifdef USE_MOUSE_ABSOLUTE
	int mouse_posx, mouse_posy;
#endif

#ifdef USE_AUTO_KEY
	FIFO* autokey_buffer;
	int autokey_phase, autokey_shift;
	uint16_t autokey_code;
	bool autokey_enabled;
#endif

#ifdef USE_DIRECTINPUT
#if DIRECTINPUT_VERSION >= 0x0800
	LPDIRECTINPUT8 lpdi;
	LPDIRECTINPUTDEVICE8 lpdikey;
//	LPDIRECTINPUTDEVICE8 lpdijoy;
#else
	LPDIRECTINPUT lpdi;
	LPDIRECTINPUTDEVICE lpdikey;
//	LPDIRECTINPUTDEVICE lpdijoy;
#endif
	uint8_t key_dik_prev[256];
#endif
	//@}

private:
	// ----------------------------------------
	// screen
	// ----------------------------------------
	/// @name screen private methods
	//@{
	void EMU_SCREEN();
	void initialize_screen();
	void release_screen();
	void release_screen_on_emu_thread();
	BOOL get_monitor_device(HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor);
	void set_vm_display_size();
	void update_screen_dc(HDC hdc);
	int  adjust_y_position(int h, int y);

#ifdef USE_DIRECT3D
	void initialize_d3device(HWND hWnd);
	HRESULT create_d3device(HWND hWnd);
	void release_d3device();
	void terminate_d3device();

	void update_screen_d3d();
	inline void copy_d3dsuf_dib(PDIRECT3DSURFACE9 suf, scrntype *buf, bool to_dib);
	void create_d3dofflinesurface();
	HRESULT reset_d3device(HWND hWnd);
	void set_d3dpresent_interval();
#endif
	//@}
	/// @name screen private static methods
	//@{
	static int qsort_by_w_h_n(const void *a, const void *b);
	static int qsort_by_n_w_h(const void *a, const void *b);
	static BOOL CALLBACK MonitorProc(HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor, LPARAM dwData);
	//@}
	/// @name screen private members
	//@{
	// screen settings
	HWND main_window_handle;
	HINSTANCE instance_handle;
	DWORD dwStyle;
	CMutex *mux_update_screen;
	bool screen_changing;

	/// vm draw area
	int screen_width, screen_height;
	int screen_width_aspect, screen_height_aspect;
	// vm draw area (not used)
//	int window_width, window_height;
	/// minimum window client size
	int min_window_width, min_window_height;
	/// display size (window client size)
	int display_width, display_height;
	/// margin in window client area (window mode only)
	VmRect display_margin;
	/// desktop size on the monitor
	int desktop_width, desktop_height, desktop_bpp;
	bool screen_size_changed;

	CSurface *dibSource;

	VmRectWH rec_video_size[2];
	VmRectWH rec_video_stretched_size;

	/// vm draw area
	/// if use screen rotation, swap width and height.
	int source_width, source_height;
	int source_width_aspect, source_height_aspect;
	int stretched_width, stretched_height;
	int stretch_pow_x, stretch_pow_y;
	int stretched_dest_x, stretched_dest_y;
	int stretched_dest_real_x, stretched_dest_real_y;
	int mixed_width, mixed_height;
	int mixed_dest_x, mixed_dest_y;
	int mixed_width_ratio, mixed_height_ratio;
	bool stretch_screen;

	// update flags
	bool first_draw_screen;
	bool first_invalidate;
	bool self_invalidate;
	bool skip_frame;
	bool skip_frame_pre;

	// screen buffer
	CSurface *dibOrigin;

#ifdef USE_SCREEN_ROTATE
	// rotate buffer
	CSurface *dibRotate;
#endif
#ifdef USE_EMU_INHERENT_SPEC
	// for message board
	CSurface *dibMixed;
#endif

#ifdef USE_SMOOTH_STRETCH
	// stretch buffer
	CSurface *dibStretch1;
	CSurface *dibStretch2;
#endif

#ifdef USE_DIRECT3D
	PDIRECT3D9			pD3D;
	D3DPRESENT_PARAMETERS	d3dpp;
	D3DDISPLAYMODE			d3ddm;
	PDIRECT3DDEVICE9	pD3Device;
	PDIRECT3DSURFACE9	pD3Dmixsuf;
	PDIRECT3DSURFACE9	pD3Dsource;
	PDIRECT3DSURFACE9	pD3Dorigin;
#ifdef USE_SCREEN_ROTATE
	PDIRECT3DSURFACE9	pD3Drotate;
#endif

	scrntype*			lpD3DBmp;
	bool                lost_d3device;
	bool                enable_direct3d;

	RECT				reD3Dmix;
	RECT				reD3Dsuf;
#endif

	// screen mode
#define DISP_DEVICE_MAX 3
	int disp_device_count;
	int monitor_count;
	struct disp_device_st {
		TCHAR    name[32];
		HMONITOR hMoni;
		RECT     re;
	} disp_device[DISP_DEVICE_MAX];

	int screen_mode_count;
	screen_mode_t screen_mode[SCREEN_MODE_MAX];

	int window_mode_count;
	window_mode_t window_mode[WINDOW_MODE_MAX];
	int window_mode_power;

	bool now_fullscreen;
	int  prev_window_mode;
	VmPoint window_dest;
	bool first_change_screen;

	// record video
	REC_VIDEO *rec_video;
	bool *now_recording_video;
	//@}

	// ----------------------------------------
	// sound
	// ----------------------------------------
	/// @name sound private methods
	//@{
	void EMU_SOUND();
	void initialize_sound(int rate, int samples, int latency);
	void initialize_sound(int rate, int samples);
	void initialize_sound();
	void release_sound();
	void release_sound_on_emu_thread();
	void start_sound();
	void end_sound();
//	void update_sound(int* extra_frames);
	DWORD update_sound_th();
	//@}
	/// @name sound private static methods
	//@{
	// callback event
	static DWORD WINAPI sound_proc(LPVOID lpParameter);
	//@}
	/// @name sound private members
	//@{
	int sound_rate, sound_samples;
	int sound_latency_half;
	bool sound_ok, sound_started, sound_finished, now_mute;

	// direct sound
	LPDIRECTSOUND lpds;
	LPDIRECTSOUNDBUFFER lpdsb, lpdsp;
//	bool first_half;
	DWORD sound_prev_time;
	// thread and event
	HANDLE sound_nt_event;
	DWORD  sound_threadid;
	HANDLE sound_thread;
	LPDIRECTSOUNDNOTIFY lpdsnt;
	CMutex *mux_sound_nt;

	// record sound
	REC_AUDIO *rec_audio;
	//@}

	// ----------------------------------------
	// floppy disk
	// ----------------------------------------
#ifdef USE_FD1
	/// @name floppy disk private methods
	//@{
	void initialize_disk_insert();
	void update_disk_insert();
	//@}
	/// @name floppy disk private members
	//@{
	typedef struct {
		_TCHAR path[_MAX_PATH];
		int offset;
		int wait_count;
		uint32_t flags;
	} disk_insert_t;
	disk_insert_t disk_insert[MAX_DRIVE];

	// d88 bank switch
	d88_file_t d88_file[MAX_DRIVE];

#define OPEN_DISK_FLAGS_READ_ONLY		0x0001
#define OPEN_DISK_FLAGS_MULTI_VOLUME	0x0010
#define OPEN_DISK_FLAGS_LAST_VOLUME		0x0008
	//@}
#endif

	// ----------------------------------------
	// media
	// ----------------------------------------
#ifdef USE_MEDIA
	/// @name media private methods
	//@{
	void initialize_media();
	void release_media();
	//@}
	/// @name media private members
	//@{
	_TCHAR media_path[MEDIA_MAX][_MAX_PATH];
	int media_cnt;
	bool media_playing;
	//@}
#endif

	// ----------------------------------------
	// timer
	// ----------------------------------------
	/// @name timer private methods
	//@{
	void update_timer();
	//@}
	/// @name timer private members
	//@{
	SYSTEMTIME sTime;
	//@}

	// ----------------------------------------
	// socket
	// ----------------------------------------
#ifdef USE_SOCKET
	/// @name socket private methods
	//@{
	void EMU_SOCKET();
	void initialize_socket();
	void release_socket();
	void update_socket();
	//@}
	/// @name socket private members
	//@{
	struct st_soc {
		SOCKET soc;
		DEVICE *device;
		bool is_tcp;
		struct sockaddr_in udpaddr;
		int socket_delay;
		char *recv_buffer;
		int recv_r_ptr, recv_w_ptr;
		CMutex *mux;
	} socs[SOCKET_MAX];
	//@}

#endif

	// ----------------------------------------
	// debugger
	// ----------------------------------------
#ifdef USE_DEBUGGER
	/// @name debugger
	//@{
	void EMU_DEBUGGER();
	void initialize_debugger();
	void release_debugger();
	HANDLE hDebuggerThread;
	debugger_thread_t debugger_thread_param;
	//@}
#endif

	// ----------------------------------------
	// misc
	// ----------------------------------------
	/// @name misc private methods
	//@{
#ifdef USE_EMU_INHERENT_SPEC
	uint64_t update_led();
#endif
	//@}
	/// @name misc private members
	//@{
	CTchar app_path;
	CTchar ini_path;

#define VM_SYSPAUSE_MASK 0x01
#define VM_SUSPEND_MASK  0x02
#define VM_POWEROFF_MASK 0x04
#define VM_USRPAUSE_MASK 0x10

	int vm_pause;

#ifdef USE_MESSAGE_BOARD
	MsgBoard *msgboard;
#endif
	//@}
//
//
//
public:
	// ----------------------------------------
	// initialize
	// ----------------------------------------
	/// @name initialize
	//@{
	EMU(const _TCHAR *new_app_path, const _TCHAR *new_ini_path);
	~EMU();
	void initialize();
	void release();
	void release_on_emu_thread();
	void set_handle(HWND hwnd, HINSTANCE hinst);
	HWND get_window() {
		return main_window_handle;
	}
	//@}

	// ----------------------------------------
	// misc
	// ----------------------------------------
	/// @name misc
	//@{
	void    application_path(char *path, int len);
	const _TCHAR *application_path();
	void    initialize_path(char *path, int len);
	const _TCHAR *initialize_path();
#if defined(_UNICODE)
	void    application_path(wchar_t *path, int len);
	void    initialize_path(wchar_t *path, int len);
#endif

	bool get_pause(int idx);
	void set_pause(int idx, bool val);
	int *get_pause_ptr(void);
	void set_gui(GUI *new_gui);
	GUI *get_gui();
	//@}

	// ----------------------------------------
	// for windows
	// ----------------------------------------
	/// @name drive virtual machine
	//@{
	int frame_interval();
	int run(int split_num);
	void update();
	bool now_skip();
	//@}
	// user interface
	/// @name control menu for ui
	//@{
	void update_config();
	void reset();
#ifdef USE_SPECIAL_RESET
	void special_reset();
	void warm_reset(int onoff);
#endif
	void change_dipswitch(int num);
#ifdef USE_POWER_OFF
	void notify_power_off();
#endif
	void change_cpu_power(int num);
#ifdef USE_EMU_INHERENT_SPEC
	void change_sync_irq();
#endif
#ifdef USE_AUTO_KEY
	bool start_auto_key(const char *str);
	bool open_auto_key(const _TCHAR *path);
	void stop_auto_key();
	bool now_auto_key() {
		return autokey_enabled;
	}
#endif
#ifdef USE_EMU_INHERENT_SPEC
	bool save_state(const _TCHAR* file_path);
	bool load_state(const _TCHAR* file_path);
#ifdef USE_KEY_RECORD
	bool play_reckey(const _TCHAR* file_path);
	bool record_reckey(const _TCHAR* file_path);
	void stop_reckey(bool stop_play = true, bool stop_record = true);
#endif
#endif
	//@}
#ifdef USE_FD1
	/// @name floppy disk menu for ui
	//@{
	bool open_disk_main(int drv, const _TCHAR* file_path, int offset, uint32_t flags);
	bool open_disk(int drv, const _TCHAR* file_path, int bank_num, int offset, uint32_t flags);
	void update_disk_info(int drv, const _TCHAR* file_path, int bank_num);
	bool open_disk_by_bank_num(int drv, const _TCHAR* file_path, int bank_num, uint32_t flags, bool multiopen);
	bool open_disk_with_sel_bank(int drv, int bank_num);
	void close_disk(int drv);
	int  change_disk(int drv);
	int  get_disk_side(int drv);
	void toggle_disk_write_protect(int drv);
	bool disk_write_protected(int drv);
	bool disk_inserted(int drv);
	bool changed_cur_bank(int drv);
	d88_file_t *get_d88_file(int drv) { return &d88_file[drv]; }
	//@}
#endif
#ifdef USE_DATAREC
	/// @name tape menu for ui
	//@{
	bool play_datarec(const _TCHAR* file_path);
	bool rec_datarec(const _TCHAR* file_path);
	void close_datarec();
	void rewind_datarec();
	void fast_forward_datarec();
	void realmode_datarec();
	bool datarec_opened(bool play_mode);
	//@}
#endif
#ifdef USE_DATAREC_BUTTON
	void push_play();
	void push_stop();
#endif
#ifdef USE_CART1
	void open_cart(int drv, const _TCHAR* file_path);
	void close_cart(int drv);
#endif
#ifdef USE_QD1
	void open_quickdisk(int drv, const _TCHAR* file_path);
	void close_quickdisk(int drv);
	void toggle_quickdisk_write_protect(int drv);
#endif
#ifdef USE_MEDIA
	void open_media(const _TCHAR* file_path);
	void close_media();
#endif
#ifdef USE_BINARY_FILE1
	void load_binary(int drv, const _TCHAR* file_path);
	void save_binary(int drv, const _TCHAR* file_path);
	void close_binary(int drv);
#endif
	/// @name screen menu for ui
	//@{
	void change_screen_mode(int mode);
	void change_stretch_screen(int num);
//	void change_cutout_screen();
	int get_window_mode_count() {
		return window_mode_count;
	}
	int get_screen_mode_count() {
		return screen_mode_count;
	}
	bool is_fullscreen() {
		return now_fullscreen;
	}
	window_mode_t *get_window_mode(int num);
	screen_mode_t *get_screen_mode(int num);
	VmRectWH *get_screen_record_size(int num);
	void change_pixel_aspect(int mode);
	int  get_pixel_aspect_count();
	int  get_pixel_aspect(int mode, int *wratio, int *hratio);
	void capture_screen();
	bool start_rec_video(int type, int fps, bool show_dialog);
	void stop_rec_video();
	void restart_rec_video();
	void record_rec_video();
	void resize_rec_video(int num);
	void change_rec_video_size(int num);
	bool now_rec_video();
	bool rec_video_enabled(int type);
	const _TCHAR **get_rec_video_codec_list(int type);
	const _TCHAR **get_rec_video_quality_list(int type);
#ifdef USE_SCANLINE
	void change_screen_scanline(int num);
#endif
#ifdef USE_EMU_INHERENT_SPEC
#ifdef USE_DIRECT3D
	void change_screen_use_direct3d(int num);
#endif
	void change_screen_afterimage(int num);
#ifdef USE_KEEPIMAGE
	void change_screen_keepimage(int num);
#endif
#endif
#ifdef USE_DIRECT3D
	bool enabled_direct3d() { return enable_direct3d; }
#endif
	//@}
	/// @name sound menu for ui
	//@{
	void start_rec_sound(int type);
	void stop_rec_sound();
	void restart_rec_sound();
	bool now_rec_sound();
	bool *get_now_rec_sound_ptr();
	bool rec_sound_enabled(int type);
	const _TCHAR **get_rec_sound_codec_list(int type);
	void set_volume(int volume);
	//@}
#ifdef USE_PRINTER
	/// @name printer menu for ui
	//@{
	bool save_printer(int dev, const _TCHAR *filename);
	void clear_printer(int dev);
	int  get_printer_buffer_size(int dev);
	uint8_t* get_printer_buffer(int dev);
//	void send_printer_data(uint8_t *datas, int len);
	void enable_printer_direct(int dev);
	bool print_printer(int dev);
	//@}
#endif
#ifdef USE_EMU_INHERENT_SPEC
	/// @name comm menu for ui
	//@{
	void enable_comm_server(int dev);
	void enable_comm_connect(int dev);
	void send_comm_telnet_command(int dev, int num);
	//@}
#endif
#ifdef USE_EMU_INHERENT_SPEC
	/// @name options menu for ui
	//@{
//	void show_led();
//	void inside_led();
//	void move_led();
//	void change_pos_led(int num);
//	int  is_shown_ledbox(int id);
	void show_message_board();
	int  is_shown_message_board();
	void change_use_joypad(int num);
	bool is_enable_joypad(int num);

	void save_keybind();

	void change_archtecture(int id, int num, bool reset);
	//@}
#endif

	//
	/// @name input device procedures for host machine
	//@{
	int  key_down_up(uint8_t type, int code, long status);
//	void key_down(int code, bool repeat);
	int  key_down(int code, bool keep_frames);
//	void key_up(int code);
	void key_up(int code, bool keep_frames);
	void key_lost_focus() {
		lost_focus = true;
	}
	int  vm_key_down(int code);
	void vm_key_up(int code);
	int  system_key_down(int code);
	void system_key_up(int code);
	bool execute_global_keys(int code, uint32_t status);
	bool release_global_keys(int code, uint32_t status);
	void post_command_message(int id);
	uint8_t translate_keysym(uint8_t type, int code, long status, int *new_code, bool *new_keep_frames = NULL);
	uint8_t translate_keysym(uint8_t type, int code, short scan_code, int *new_code, bool *new_keep_frames = NULL);

#ifdef USE_BUTTON
	void press_button(int num);
#endif
	void reset_joystick();
	void update_joystick();

	inline void update_mouse();
	void enable_mouse(int mode);
	void disable_mouse(int mode);
	void toggle_mouse();
	bool get_mouse_enabled();

#ifdef USE_AUTO_KEY
	void parse_auto_key(const char *buf, int size, int &prev_code, int &prev_kana);
#endif
	void update_autokey();
	void enable_joystick();
	//@}
	/// @name screen device procedures for host machine
	//@{
//	int get_window_width(int mode);
//	int get_window_height(int mode);
	void set_display_size(int width, int height, int power, bool window_mode);
//	void change_screen_size(int sw, int sh, int swa, int sha, int ww, int wh);
	void draw_screen();
	void fill_gray();
	void mix_screen();
	void update_screen(HDC hdc);
	void skip_screen(bool val) {
		skip_frame = val;
	}
	void need_update_screen();

	void enum_window_mode(int max_width, int max_height);
	void enum_screen_mode();
	void find_screen_mode();
	void set_client_pos(int dest_x, int dest_y, int client_width, int client_height, UINT flags);
	void set_window(int mode, int cur_width, int cur_height);
	void set_desktop_size(int width, int height, int bpp) {
		desktop_width = width;
		desktop_height = height;
		desktop_bpp = bpp;
	}
	void set_display_margin(int left, int top, int right, int bottom) {
		display_margin.left = left;
		display_margin.top = top;
		display_margin.right = right;
		display_margin.bottom = bottom;
	}
	void set_window_style(int x, int y, DWORD style) {
		window_dest.x = x;
		window_dest.y = y;
		dwStyle = style;
	}
	void lock_screen();
	void unlock_screen();
#ifdef USE_BITMAP
	void reload_bitmap() {
		first_invalidate = true;
	}
#endif
#ifdef USE_DIRECT3D
	PDIRECT3DDEVICE9 get_d3device() { return pD3Device; }
#endif
	//@}
	/// @name sound device procedures for host machine
	//@{
	void mute_sound(bool mute);
	uint32_t adjust_sound_pos(uint32_t msec);
	//@}
#ifdef USE_SOCKET
	/// @name socket device procedures for host machine
	//@{
	SOCKET get_socket(int ch) {
		return socs[ch].soc;
	}
	void socket_connected(int ch);
	void socket_disconnected(int ch);
	void socket_writeable(int ch);
	void socket_readable(int ch);
	void socket_accept(int ch);
	void send_data(int ch);
	void recv_data(int ch);
	//@}
#endif

#ifdef USE_DEBUGGER
	/// @name debugger procedures for host machine
	//@{
	void open_debugger(int cpu_index);
	void close_debugger();
	bool debugger_enabled(int cpu_index);
	void debugger_terminal_accepted(DEBUGGER *debugger);
	bool now_debugging;

	bool debugger_save_image(int width, int height, CSurface *surface);
	//@}
#endif

	// ----------------------------------------
	// for virtual machine
	// ----------------------------------------

	// power off
	void power_off();
	void toggle_power_on_off();

	/// @name input device for vm
	//@{
	uint8_t* key_buffer() {
		return key_status;
	}
#ifdef USE_JOYSTICK
	uint32_t* joy_buffer() {
		return joy_status;
	}
#endif
	int* mouse_buffer() {
		return mouse_status;
	}
	int* get_key_mod_ptr() {
		return &key_mod;
	}
	void recognized_key(uint16_t key_code);
	void set_vm_key_buffer(uint8_t *buffer, int size) {
		vm_key_status = buffer;
		vm_key_status_size = size;
	}
	void get_vm_key_buffer(uint8_t **buffer, int *size) {
		*buffer = vm_key_status;
		*size = vm_key_status_size;
	}
	//@}
	/// @name screen for vm
	//@{
	scrntype* screen_buffer(int y);
	int screen_buffer_offset();
	void get_rgbformat(uint32_t *r_mask, uint32_t *g_mask, uint32_t *b_mask, uint8_t *r_shift, uint8_t *g_shift, uint8_t *b_shift);
	void get_rgbaformat(uint32_t *r_mask, uint32_t *g_mask, uint32_t *b_mask, uint32_t *a_mask, uint8_t *r_shift, uint8_t *g_shift, uint8_t *b_shift, uint8_t *a_shift);
	void get_rgbcolor(scrntype pixel, uint8_t *r, uint8_t *g, uint8_t *b);
	scrntype map_rgbcolor(uint8_t r, uint8_t g, uint8_t b);
	bool now_skip_frame();
	//@}
	/// @name sound for vm
	//@{
//	void record_rec_sound(uint8_t *buffer, int samples);
//	void record_rec_sound(int16_t *buffer, int samples);
	void record_rec_sound(int32_t *buffer, int samples);
	void lock_sound_buffer();
	void unlock_sound_buffer();
	//@}
	/// @name timer for vm
	//@{
	void get_timer(int time[]);
	//@}
#ifdef USE_MEDIA
	/// @name media for vm
	//@{
	int media_count();
	void play_media(int trk);
	void stop_media();
	//@}
#endif
#ifdef USE_SOCKET
	/// @name socket for vm
	//@{
	bool init_socket_tcp(int ch, DEVICE *dev, bool server = false);
	bool init_socket_udp(int ch, DEVICE *dev, bool server = false);
	bool connect_socket(int ch, uint32_t ipaddr, int port, bool server = false);
	bool connect_socket(int ch, const _TCHAR *hostname, int port, bool server = false);
	bool get_ipaddr(const _TCHAR *hostname, int port, uint32_t *ipaddr);
	void disconnect_socket(int ch);
	int  get_socket_channel();
	bool listen_socket(int ch);
	void send_data_tcp(int ch);
	void send_data_udp(int ch, uint32_t ipaddr, int port);
	void send_data_udp(int ch, const _TCHAR *hostname, int port);
	//@}
#endif
	/// @name load rom image
	//@{
	bool load_data_from_file(const _TCHAR *file_path, const _TCHAR *file_name
		, uint8_t *data, size_t size
		, const uint8_t *first_data = NULL, size_t first_data_size = 0
		, const uint8_t *last_data = NULL,  size_t last_data_size = 0);
	//@}
#ifdef USE_EMU_INHERENT_SPEC
	/// @name send message to ui from vm
	//@{
	void update_ui(int flags);
	//@}
//	void send_comm_data(uint8_t *data, size_t size);
//	void recv_comm_data(uint8_t *data, size_t size);
	/// @name misc
	//@{
	int parami[VM::ParamiUnknown];
	void *paramv[VM::ParamvUnknown];
	int get_parami(int id);
	void set_parami(int id, int val);
	void *get_paramv(int id);
	void set_paramv(int id, void *val);
	void update_params();
	//@}
#endif
	/// @name for debug
	//@{
	uint64_t get_current_clock();
	void sleep(uint32_t ms);
	//@}

	/// @name disp message methods
	//@{
	void send_log_message(int level, const _TCHAR *levelstr, const _TCHAR *msg);

	void out_msg(const _TCHAR* msg, bool set, int sec);
	void out_msg(const _TCHAR* msg, bool set = true);
	void out_info(const _TCHAR* msg, bool set, int sec);
	void out_info(const _TCHAR* msg, bool set = true);
	void out_infof(const _TCHAR* format, ...);

#ifdef USE_MESSAGE_BOARD
	MsgBoard *get_msgboard() { return msgboard; }
#endif
	//@}
};

#endif
