/** @file win_vkeyboard.cpp

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@author Sasaji
	@date   2016.12.21 -

	@brief [ virtual keyboard ]
*/

#include "win_vkeyboard.h"
#include "../../emu.h"
#include "../../res/resource.h"

extern EMU *emu;

namespace Vkbd {

//
// for windows
//
static const int csBitmapIds[BITMAPIDS_END] = {
	IDB_VKEYBOARD,
	IDB_LED_PARTS,
	IDB_VKEY_MODE,
	IDB_VKEY_BREAK,
	IDB_VKEY_POWER
};

VKeyboard::VKeyboard(HINSTANCE hInst, HWND hWnd) : Base()
{
	hInstance = hInst;
	hParent = hWnd;
	hVkbd = NULL;
	pSurface = NULL;
	hBrush = NULL;

	for(int i=0; i<BITMAPIDS_END; i++) {
		pBitmaps[i] = NULL;
	}
}

VKeyboard::~VKeyboard()
{
	for(int i=0; i<BITMAPIDS_END; i++) {
		delete pBitmaps[i];
	}
}

void VKeyboard::Show(bool show)
{
	if (!hVkbd) return;

	::ShowWindow(hVkbd, show ? SW_SHOWNORMAL : SW_HIDE);
}

void VKeyboard::Create()
{
	if (hVkbd) return;

	hVkbd = ::CreateDialogParam(hInstance, MAKEINTRESOURCE(IDD_VKEYBOARD), hParent, Proc, (LPARAM)this);
	if (hVkbd) {
		load_bitmap();
		create_surface();
		adjust_window_size();
		set_dist();
		Base::Create();
	}
}

void VKeyboard::Close()
{
	if (!hVkbd) return;

	::EndDialog(hVkbd, 0);
	hVkbd = NULL;

	unload_bitmap();

	Base::Close();
}

INT_PTR CALLBACK VKeyboard::Proc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	UNREFERENCED_PARAMETER(lParam);
	PAINTSTRUCT ps;
	HDC hdc;
	VKeyboard *myObj = (VKeyboard *)GetWindowLongPtr(hDlg, DWLP_USER);

	switch (message)
	{
	case WM_PAINT:
		hdc = BeginPaint(hDlg, &ps);
		myObj->paint_window(hdc, &ps.rcPaint);
		EndPaint(hDlg, &ps);
		return (INT_PTR)FALSE;
	case WM_INITDIALOG:
		SetWindowLongPtr(hDlg, DWLP_USER, lParam);
		myObj = (VKeyboard *)lParam;
		myObj->init_dialog(hDlg);
		return (INT_PTR)TRUE;
	case WM_LBUTTONDOWN:
	case WM_LBUTTONDBLCLK:
		myObj->MouseDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		return (INT_PTR)FALSE;
	case WM_LBUTTONUP:
		myObj->MouseUp();
		return (INT_PTR)FALSE;
	case WM_SYSKEYDOWN:
	case WM_SYSKEYUP:
#ifdef USE_ALT_F10_KEY
		return (INT_PTR)FALSE;	// not activate menu when hit ALT/F10
#endif
	case WM_KEYDOWN:
	case WM_KEYUP:
		if (emu) {
			emu->key_down_up(message & 1, LOBYTE(wParam), (long)lParam);
		}
		return (INT_PTR)FALSE;
	case WM_COMMAND:
		switch(LOWORD(wParam)) {
		case IDOK:
		case IDCANCEL:
			myObj->Close();
			return (INT_PTR)TRUE;
		}
		break;
	case WM_CLOSE:
		myObj->Close();
		return (INT_PTR)TRUE;
//	case WM_DESTROY:
//		break;
	}
	return (INT_PTR)FALSE;
}

void VKeyboard::load_bitmap()
{
	// set bitmaps
	HDC hdc = ::GetDC(hVkbd);
	for(int i=0; i<BITMAPIDS_END; i++) {
		pBitmaps[i] = new CBitmap(hInstance, hdc, csBitmapIds[i], 1);
	}
	::ReleaseDC(hVkbd, hdc);

	// brush
#ifdef _BML3MK5
	hBrush = ::CreateSolidBrush(RGB(0x80,0x80,0x80));
#else
	hBrush = ::CreateSolidBrush(RGB(0x30,0x30,0x30));
#endif
}

bool VKeyboard::create_surface()
{
	if (pSurface) return true;
	if (!hVkbd) return false;

	CBitmap *bmp = pBitmaps[BITMAPIDS_BASE];
	HDC hdc = ::GetDC(hVkbd);
	pSurface = new CSurface(hdc, bmp->Width(), bmp->Height());
	::BitBlt(pSurface->GetDC(), 0, 0, pSurface->Width(), pSurface->Height(), bmp->GetDC(), 0, 0, SRCCOPY);
	::ReleaseDC(hVkbd, hdc);
	return true;
}

void VKeyboard::unload_bitmap()
{
	if (pSurface) {
		delete pSurface;
		pSurface = NULL;
	}

	// release bitmaps
	for(int i=0; i<BITMAPIDS_END; i++) {
		delete pBitmaps[i];
		pBitmaps[i] = NULL;
	}

	if (hBrush) {
		::DeleteObject(hBrush);
		hBrush = NULL;
	}
}

void VKeyboard::adjust_window_size()
{
	if (!hVkbd || !pSurface) return;

	WINDOWINFO wi;

	// calc client size in the window
	GetWindowInfo(hVkbd, &wi);
	int width = wi.rcClient.left - wi.rcWindow.left + wi.rcWindow.right - wi.rcClient.right + pSurface->Width();
	int height = wi.rcClient.top - wi.rcWindow.top + wi.rcWindow.bottom - wi.rcClient.bottom + pSurface->Height();
	::SetWindowPos(hVkbd, HWND_TOPMOST, 0, 0, width, height, SWP_NOMOVE | SWP_NOZORDER);
}

void VKeyboard::set_dist()
{
	if (!hVkbd || !hParent) return;

	WINDOWINFO wip;
	WINDOWINFO wi;

	HDC hdcScr = ::GetDC(NULL);
	int desktop_width = ::GetDeviceCaps(hdcScr, HORZRES);
	int desktop_height = ::GetDeviceCaps(hdcScr, VERTRES);
	::ReleaseDC(NULL, hdcScr);

	::GetWindowInfo(hParent, &wip);
	::GetWindowInfo(hVkbd, &wi);

	int wp = wip.rcWindow.right - wip.rcWindow.left;
	int hp = wip.rcWindow.bottom - wip.rcWindow.top;
	int w = wi.rcWindow.right - wi.rcWindow.left;
	int h = wi.rcWindow.bottom - wi.rcWindow.top;

	int x = (wp - w) / 2 + wip.rcWindow.left;
	int y = wip.rcWindow.bottom;

	if (y + h > desktop_height) {
		y = (desktop_height - h);
	}

	::SetWindowPos(hVkbd, HWND_TOPMOST, x, y, 1, 1, SWP_NOSIZE | SWP_NOZORDER);
}

void VKeyboard::need_update_window(PressedInfo_t *info, bool onoff)
{
	if (!hVkbd) return;

	const Bitmap_t *bp = NULL;
	
	Rect_t dstrect = info->re;
	
	if (onoff) {
		// on
		if (info->parts_num >= 0) {
			bp = &cBmpParts[info->parts_num];
			blit_surface(pBitmaps[bp->idx], bp->x, bp->y, &dstrect);
		} else {
			fill_rect(&dstrect);
			dstrect.top += 3;
			blit_surface(pBitmaps[BITMAPIDS_BASE], info->re.left, info->re.top,	&dstrect);
		}
	} else {
		// off
		blit_surface(pBitmaps[BITMAPIDS_BASE], dstrect.left, dstrect.top, &dstrect);
	}

	RECT re = { info->re.left, info->re.top, info->re.right, info->re.bottom };
	::InvalidateRect(hVkbd, &re, TRUE);
}

void VKeyboard::update_window()
{
	if (!hVkbd) return;

	::UpdateWindow(hVkbd);
}

void VKeyboard::paint_window(HDC hdc, RECT *re)
{
	if (!pSurface) return;

	::BitBlt(hdc, re->left, re->top, re->right - re->left, re->bottom - re->top, pSurface->GetDC(), re->left, re->top, SRCCOPY);
}

void VKeyboard::fill_rect(Rect_t *re)
{
	if (!pSurface) return;

	RECT wre = { re->left, re->top, re->right, re->bottom };
	::FillRect(pSurface->GetDC(), &wre, hBrush);
}

bool VKeyboard::blit_surface(CBitmap *pSrcBmp, short src_x, short src_y, Rect_t *dst_re)
{
	if (!pSurface || !pSrcBmp) return false;

	::BitBlt(pSurface->GetDC(), dst_re->left, dst_re->top, dst_re->right - dst_re->left, dst_re->bottom - dst_re->top,
		pSrcBmp->GetDC(), src_x, src_y, SRCCOPY);

	return true;
}

void VKeyboard::init_dialog(HWND hDlg)
{
	// disable ime
	ImmAssociateContext(hDlg, NULL);

	// set icon on sysmenu
	HICON hIcon = ::LoadIconA(hInstance, MAKEINTRESOURCE(IDI_ICON1));
	::SendMessageA(hDlg, WM_SETICON, NULL, (LPARAM)hIcon);
}

} /* namespace Vkbd */
