/** @file rec_video.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.05.18 -

	@brief [ record video ]
*/

#include <Windows.h>
#include "rec_video.h"
#include "../emu.h"
#include "../csurface.h"

#ifdef USE_REC_VIDEO
#ifdef USE_REC_VIDEO_VFW
#include "vfw/vfw_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_MMF
#include "mmf/mmf_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_FFMPEG
#include "ffmpeg/ffm_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_WIN
#include "windows/win_rec_video.h"
#endif
#endif /* USE_REC_VIDEO */

REC_VIDEO::REC_VIDEO(EMU *new_emu)
{
	emu = new_emu;
	now_recording = false;
	rec_type = 0;
	rec_fps = 0;
	rec_count = 0;
	rec_denomi = 1;
	memset(rec_path, 0, sizeof(rec_path));

	rec_surface = new CSurface();

#ifdef USE_REC_VIDEO
#ifdef USE_REC_VIDEO_VFW
	vfwvideo = new VFW_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_MMF
	mmfvideo = new MMF_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	ffmvideo = new FFM_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_WIN
	winvideo = new WIN_REC_VIDEO(new_emu, this);
#endif
#endif /* USE_REC_VIDEO */
}

REC_VIDEO::~REC_VIDEO()
{
#ifdef USE_REC_VIDEO
#ifdef USE_REC_VIDEO_FFMPEG
	delete ffmvideo;
#endif
#ifdef USE_REC_VIDEO_MMF
	delete mmfvideo;
#endif
#ifdef USE_REC_VIDEO_VFW
	delete vfwvideo;
#endif
#ifdef USE_REC_VIDEO_WIN
	delete winvideo;
#endif
#endif /* USE_REC_VIDEO */
	delete rec_surface;
}

bool REC_VIDEO::IsEnabled(int type)
{
	bool rc = false;
#ifdef USE_REC_VIDEO
	switch(type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		rc = true;
		break;
#endif
#ifdef USE_REC_VIDEO_MMF
	case RECORD_VIDEO_TYPE_MMF:
		rc = mmfvideo->IsEnabled();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		rc = ffmvideo->IsEnabled();
		break;
#endif
	default:
		break;
	}
#endif
	return rc;
}

const _TCHAR **REC_VIDEO::GetCodecList(int type)
{
	const _TCHAR **list = NULL;
#ifdef USE_REC_VIDEO
	switch(type) {
#ifdef USE_REC_VIDEO_MMF
	case RECORD_VIDEO_TYPE_MMF:
		list = mmfvideo->GetCodecList();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		list = ffmvideo->GetCodecList();
		break;
#endif
	default:
		break;
	}
#endif
	return list;
}

const _TCHAR **REC_VIDEO::GetQualityList(int type)
{
	const _TCHAR **list = NULL;
#ifdef USE_REC_VIDEO
	switch(type) {
#ifdef USE_REC_VIDEO_MMF
	case RECORD_VIDEO_TYPE_MMF:
		list = mmfvideo->GetQualityList();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		list = ffmvideo->GetQualityList();
		break;
#endif
	default:
		break;
	}
#endif
	return list;
}

bool REC_VIDEO::Start(int type, int fps, VmRectWH *srcrect, CSurface *srcsurface, bool show_dialog)
{
#ifdef USE_REC_VIDEO
	if (type > 0) {
		rec_type = type;
	}
	if(fps > 0) {
		rec_fps = fps;
		rec_count = 0;
		rec_denomi = 60 / fps;
	}

	if (rec_type <= 0 || rec_fps <= 0) {
		return false;
	}

	CreateFileName(rec_path);

	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		if (srcrect != NULL) CreateSurface(srcrect->w, srcrect->h);
		now_recording = vfwvideo->Start(rec_path, rec_fps, srcrect, rec_surface, show_dialog);
		break;
#endif
#ifdef USE_REC_VIDEO_MMF
	case RECORD_VIDEO_TYPE_MMF:
		if (srcrect != NULL) CreateSurface(srcrect->w, srcrect->h);
		now_recording = mmfvideo->Start(rec_path, rec_fps, srcrect, rec_surface, show_dialog);
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		// height set minus, so create the flipped surface
		if (srcrect != NULL) CreateSurface(srcrect->w, -srcrect->h);
		now_recording = ffmvideo->Start(rec_path, rec_fps, srcrect, rec_surface, show_dialog);
		break;
#endif
	}
#endif
	return now_recording;
}

void REC_VIDEO::Stop()
{
	if (!now_recording) return;
#ifdef USE_REC_VIDEO
	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		vfwvideo->Stop();
		break;
#endif
#ifdef USE_REC_VIDEO_MMF
	case RECORD_VIDEO_TYPE_MMF:
		mmfvideo->Stop();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		ffmvideo->Stop();
		break;
#endif
	}
#endif
	now_recording = false;
}

bool REC_VIDEO::Restart()
{
	if (!now_recording) return false;
#ifdef USE_REC_VIDEO
	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		now_recording = vfwvideo->Restart();
		break;
#endif
#ifdef USE_REC_VIDEO_MMF
	case RECORD_VIDEO_TYPE_MMF:
		now_recording = mmfvideo->Restart();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		now_recording = ffmvideo->Restart();
		break;
#endif
	}
#endif
	return now_recording;
}

bool REC_VIDEO::Record(VmRectWH *srcrect, CSurface *srcsurface, VmRectWH *dstrect)
{
	if (!now_recording) return false;

#ifdef USE_REC_VIDEO
	if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
		BitBlt(rec_surface->GetDC(), 0, 0, dstrect->w, dstrect->h, srcsurface->GetDC(), srcrect->x, srcrect->y, SRCCOPY);
	} else {
		StretchBlt(rec_surface->GetDC(), 0, 0, dstrect->w, dstrect->h, srcsurface->GetDC(), srcrect->x, srcrect->y, srcrect->w, srcrect->h, SRCCOPY);
	}

	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		now_recording = vfwvideo->Record();
		break;
#endif
#ifdef USE_REC_VIDEO_MMF
	case RECORD_VIDEO_TYPE_MMF:
		now_recording = mmfvideo->Record();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		now_recording = ffmvideo->Record();
		break;
#endif
	}
#endif
	return now_recording;
}

bool REC_VIDEO::IsRecordFrame()
{
	if (!now_recording) return false;

	bool rc = (rec_count == 0);
	rec_count = ((rec_count + 1) % rec_denomi);
	return rc;
}

static const _TCHAR *capture_file_ext[] = {
	_T(".bmp"),
	_T(".png"),
	NULL
};

bool REC_VIDEO::Capture(int type, VmRectWH *srcrect, CSurface *srcsurface, VmRectWH *dstrect)
{
	if (srcrect == NULL || srcsurface == NULL) {
		return false;
	}

	bool rc = false;
	CSurface bmpsurface;
	_TCHAR file_path[_MAX_PATH];
	CreateFileName(file_path);

#if !(defined(USE_REC_VIDEO_WIN) || defined(USE_REC_VIDEO_COCOA) || defined(USE_REC_VIDEO_LIBPNG))
	config.capture_type = 0;
#endif
	if (1 < config.capture_type) {
		config.capture_type = 0;
	}

	// add ext
	_tcscat(file_path, capture_file_ext[config.capture_type]);

	// create surface for capture
	HDC hdc = GetDC(emu->get_window());
	switch(config.capture_type) {
	case 1:	// PNG
#if defined(USE_REC_VIDEO_WIN)
		rc = bmpsurface.Create(hdc, dstrect->w, dstrect->h * -1); // flipped
#endif
		break;
	default: // BMP
		rc = bmpsurface.Create(hdc, dstrect->w, dstrect->h);
		break;
	}
	ReleaseDC(emu->get_window(), hdc);
	if (!rc) {
		emu->out_log(LOG_ERROR,_T("Capture screen failed. Cannot create dib."));
		return false;
	}

	// copy surface
	if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
		BitBlt(bmpsurface.GetDC(), 0, 0, dstrect->w, dstrect->h, srcsurface->GetDC(), srcrect->x, srcrect->y, SRCCOPY);
	} else {
		StretchBlt(bmpsurface.GetDC(), 0, 0, dstrect->w, dstrect->h, srcsurface->GetDC(), srcrect->x, srcrect->y, srcrect->w, srcrect->h, SRCCOPY);
	}

	switch(config.capture_type) {
	case 1:	// PNG
	{
#if defined(USE_REC_VIDEO_WIN)
		CTchar cfile_path(file_path);
		rc = winvideo->Capture(type, &bmpsurface, cfile_path.GetW());
#endif
		break;
	}
	default: // BMP
	{
		// save surface
		rc = SaveBMP(&bmpsurface, file_path);
		break;
	}
	}
	if (rc) {
		if (type == CAPTURE_SCREEN_TYPE) {
			emu->out_log(LOG_INFO, _("Screen was saved successfully."));
		}
	} else {
		emu->out_log(LOG_ERROR, _T("Cannot save screen captured file."));
	}

	return rc;
}

bool REC_VIDEO::SaveBMP(CSurface *surface, const _TCHAR *file_name)
{
	DWORD dwSize;
	HANDLE hFile = INVALID_HANDLE_VALUE;
	BITMAPFILEHEADER bmFileHeader = { (WORD)(TEXT('B') | TEXT('M') << 8) };

	bmFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	bmFileHeader.bfSize = bmFileHeader.bfOffBits + surface->GetBufferSize();

	hFile = CreateFile(file_name, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		emu->out_syserrlog(LOG_ERROR,GetLastError(),_T("Capture screen failed."));
		return false;
	}
	WriteFile(hFile, &bmFileHeader, sizeof(BITMAPFILEHEADER), &dwSize, NULL);
	WriteFile(hFile, surface->GetHeader(), sizeof(BITMAPINFOHEADER), &dwSize, NULL);
	WriteFile(hFile, surface->GetBuffer(), surface->GetBufferSize(), &dwSize, NULL);
	CloseHandle(hFile);

	return true;
}

void REC_VIDEO::CreateFileName(_TCHAR *file_path)
{
	int tim[8];
	const _TCHAR *app_path;

	app_path = config.snapshot_path.Length() > 0 ? config.snapshot_path : emu->application_path();
	emu->get_timer(tim);
	_stprintf(file_path, _T("%s%04d-%02d-%02d_%02d-%02d-%02d"), app_path, tim[0], tim[1], tim[2], tim[4], tim[5], tim[6]);
}

void REC_VIDEO::CreateSurface(int w, int h)
{
	HDC hdc = GetDC(emu->get_window());
	rec_surface->Create(hdc, w, h);
	ReleaseDC(emu->get_window(), hdc);
}
