/** @file win32_input_keysym.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2014.04.01

	@brief [ win32 input keysym ]
*/

#include "emu.h"
#include "vm/vm.h"
#include "gui/gui.h"
#include "win32_input.h"

/// notify accel key
int EMU::system_key_down(int code)
{
	if (vm) {
		code = translate_global_key(code);
		vm->system_key_down(code);
	}
	return code;
}
void EMU::system_key_up(int code)
{
	if (vm) {
		code = translate_global_key(code);
		vm->system_key_up(code);
	}
}

/// @param[in] type bit0:0:down 1:up  bit1:0:WM_KEYDOWN/UP 1:direct_input  bit2:1:avail dinput
uint8_t EMU::translate_keysym(uint8_t type, int code, long status, int *new_code, bool *new_keep_frames)
{
	short scan_code = (short)((status & 0x1ff0000) >> 16);
	return translate_keysym(type, code, scan_code, new_code, new_keep_frames);
}

uint8_t EMU::translate_keysym(uint8_t type, int code, short scan_code, int *new_code, bool *new_keep_frames)
{
	int n_code = 0;
	bool n_keep_frames = false;

#ifdef USE_DIRECTINPUT
	if (!(type & 2)) {
		if (type & 4) {
			if (code < 256) {
				n_code = vkkey2keycode[code];
				n_keep_frames = ((n_code & 0x8000) != 0);
				n_code &= 0xff;
				if (n_code == 0) return type;
			}
		} else {
#endif
			if(code == VK_SHIFT) {
				if(scan_code == 0x36 || GetKeyState(VK_RSHIFT) & 0x8000) n_code = KEYCODE_RSHIFT;
				else n_code = KEYCODE_LSHIFT;
			}
			else if(code == VK_CONTROL) {
				if((scan_code & 0x100) || GetKeyState(VK_RCONTROL) & 0x8000) n_code = KEYCODE_RCTRL;
				else n_code = KEYCODE_LCTRL;
			}
			else if(code == VK_MENU) {
				if((scan_code & 0x100) || GetKeyState(VK_RMENU) & 0x8000) n_code = KEYCODE_RALT;
				else n_code = KEYCODE_LALT;
			}
			else if(code == VK_RETURN) {
				if (scan_code & 0x100) n_code = KEYCODE_KP_ENTER;
				else n_code = KEYCODE_RETURN;
			}
			else if (code < 256) {
				n_code = vkkey2keycode[code];
				n_keep_frames = ((n_code & 0x8000) != 0);
				n_code &= 0xff;
			}
#ifdef USE_DIRECTINPUT
		}
#endif
		// convert numpad keys
		if (scan_code >= 0x47 && scan_code <= 0x53) {
			if (scancode2keycode47[scan_code - 0x47] != 0) n_code = scancode2keycode47[scan_code - 0x47];
		}
		else if (scan_code >= 0x70 && scan_code <= 0x7f) {
			if (scancode2keycode70[scan_code - 0x70] != 0) n_code = scancode2keycode70[scan_code - 0x70];
		}
		else if (scan_code == 0x29) {
			// code is set either Hankaku or Kanji in JP
			n_code = KEYCODE_GRAVE;
			n_keep_frames = true;
		}
#ifdef USE_DIRECTINPUT
	} else {
		n_code = dikey2keycode[code];
		n_keep_frames = ((n_code & 0x8000) != 0);
		n_code &= 0xff;
	}
#endif

	if (n_code == 0) {
		n_code = code | 0x100;
	}
	if (new_code) *new_code = n_code;
	if (new_keep_frames) *new_keep_frames = n_keep_frames;

//	out_debugf("keytrans: %d %02x->%02x %04x keep:%s", type, code, n_code, scan_code, n_keep_frames ? "true" : "false");

	return (type & 1);
}

/// execute accel key
bool EMU::execute_global_keys(int code, uint32_t status)
{
	if (gui) {
		code = translate_global_key(code);
		return gui->ExecuteGlobalKeys(code, status);
	}
	return false;
}

/// release accel key
bool EMU::release_global_keys(int code, uint32_t status)
{
	if (gui) {
		code = translate_global_key(code);
		return gui->ReleaseGlobalKeys(code, status);
	}
	return false;
}

/// translate
int EMU::translate_global_key(int code)
{
	if (code == KEYCODE_LCTRL || code == KEYCODE_RCTRL) {
		code = GLOBALKEY_CONTROL;
	} else if (code == KEYCODE_RETURN) {
		code = GLOBALKEY_RETURN;
	} else if (code == KEYCODE_0) {
		code = GLOBALKEY_0;
	} else if (KEYCODE_1 <= code && code <= KEYCODE_9) {
		code = code + GLOBALKEY_1 - KEYCODE_1;
	} else if (KEYCODE_A <= code && code <= KEYCODE_Z) {
		code = code + GLOBALKEY_A - KEYCODE_A;
	} else if (KEYCODE_F1 <= code && code <= KEYCODE_F12) {
		code = code + GLOBALKEY_F1 - KEYCODE_F1;
	} else if (KEYCODE_F13 <= code && code <= KEYCODE_F15) {
		code = code + GLOBALKEY_F13 - KEYCODE_F13;
	}
	return code;
}

/// post message to main thread
void EMU::post_command_message(int id)
{
	if (gui) gui->PostCommandMessage(id);
}