/** @file win32_screen.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya, Sasaji
	@date   2006.08.18 -

	@brief [ win32 screen ]
*/

#include "emu.h"
#include "vm/vm.h"
#include "gui/gui.h"
#include "config.h"
#ifdef USE_MESSAGE_BOARD
#include "msgboard.h"
#endif
#include "csurface.h"
#include "video/rec_video.h"

static const SIZE window_proposal[] = {
	{640,400},{640,480},{800,600},{1024,768},{0,0}
};

void EMU::EMU_SCREEN()
{
	screen_width = SCREEN_WIDTH;
	screen_height = SCREEN_HEIGHT;
	screen_width_aspect = SCREEN_WIDTH_ASPECT;
	screen_height_aspect = SCREEN_HEIGHT_ASPECT;
//	window_width = WINDOW_WIDTH;
//	window_height = WINDOW_HEIGHT;
	min_window_width = MIN_WINDOW_WIDTH;
	min_window_height = MIN_WINDOW_HEIGHT;
	screen_size_changed = true;
	desktop_width = MIN_WINDOW_WIDTH;
	desktop_height = MIN_WINDOW_HEIGHT;
	desktop_bpp = 24;
	display_margin.left = 0;
	display_margin.top = 0;
	display_margin.right = 0;
	display_margin.bottom = 0;
	dwStyle = 0;

	source_width = screen_width;
	source_height = screen_height;
	source_width_aspect = source_height_aspect = -1;
	stretch_pow_x = stretch_pow_y = -1;
//	display_stretched_width = display_stretched_height = 1;
	stretch_screen = false;
	stretched_width = SCREEN_WIDTH;
	stretched_height = SCREEN_HEIGHT;
	stretched_dest_x = stretched_dest_y = 0;
	stretched_dest_real_x = stretched_dest_real_y = 0;
	mixed_width = source_width;
	mixed_height = source_height;
	mixed_dest_x = mixed_dest_y = 0;
	mixed_width_ratio = mixed_height_ratio = 1;

	dibOrigin = NULL;
	dibSource = NULL;
	dibMixed = NULL;

#ifdef USE_SMOOTH_STRETCH
	dibStretch1 = NULL;
	dibStretch2 = NULL;
#endif

	// screen mode
	disp_device_count = 0;
	monitor_count = 0;
	screen_mode_count = 0;
	window_mode_count = 0;
	window_mode_power = 10;
	now_fullscreen = false;
	prev_window_mode = 0;
	window_dest.x = 0;
	window_dest.y = 0;
	first_change_screen = true;

#ifdef USE_DIRECT3D
	pD3D = NULL;
	pD3Device = NULL;
	pD3Dmixsuf = NULL;
	pD3Dorigin = NULL;
	pD3Dsource = NULL;
#endif

	// initialize video recording
	rec_video_size[0].w = MIN_WINDOW_WIDTH;
	rec_video_size[0].h = MIN_WINDOW_HEIGHT;
	rec_video_size[0].x = (MAX_WINDOW_WIDTH - MIN_WINDOW_WIDTH) / 2;
	rec_video_size[0].y = (MAX_WINDOW_HEIGHT - MIN_WINDOW_HEIGHT) / 2;
	rec_video_size[1].w = MAX_WINDOW_WIDTH;
	rec_video_size[1].h = MAX_WINDOW_HEIGHT;
	rec_video_size[1].x = 0;
	rec_video_size[1].y = 0;
	rec_video_stretched_size = rec_video_size[0];

	rec_video = new REC_VIDEO(this);
	now_recording_video = rec_video->GetNowRecordingPtr();

	// initialize update flags
	first_draw_screen = false;
	first_invalidate = self_invalidate = false;
	skip_frame = false;

	// for lock screen
	mux_update_screen = new CMutex();
	screen_changing = false;
}

void EMU::initialize_screen()
{
	out_debug(_T("EMU::initialize_screen"));

	// create dib sections
	HDC hdc = GetDC(main_window_handle);
	dibOrigin = new CSurface(hdc, screen_width, screen_height);
#ifdef USE_SCREEN_ROTATE
	dibRotate = new CSurface(hdc, screen_height, screen_width);
#endif
#ifdef USE_EMU_INHERENT_SPEC
	// for message board
	dibMixed = new CSurface(hdc, screen_width, screen_height);
#endif
#ifdef USE_SMOOTH_STRETCH
	dibStretch1 = new CSurface();
	dibStretch2 = new CSurface();
#endif
#ifdef USE_DIRECT3D
	initialize_d3device(main_window_handle);
	create_d3device(main_window_handle);
	create_d3dofflinesurface();
	if (pD3Device != NULL) {
		pD3Device->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,0,0), 0.0, 0);
	}
#endif
	change_pixel_aspect(config.pixel_aspect);
	change_rec_video_size(config.screen_video_size);

#ifdef USE_MESSAGE_BOARD
	msgboard = new MsgBoard(main_window_handle, this);
	if (msgboard) {
		msgboard->InitScreen(hdc, screen_width, screen_height);
		msgboard->SetVisible(FLG_SHOWMSGBOARD ? true : false);
	}
#endif
	if (gui) {
		gui->CreateLedBox();
	}
	ReleaseDC(main_window_handle, hdc);
}

void EMU::release_screen()
{
	// release dib sections
	delete dibOrigin;
	dibOrigin = NULL;
#ifdef USE_SCREEN_ROTATE
	delete dibRotate;
	dibRotate = NULL;
#endif
#ifdef USE_EMU_INHERENT_SPEC
	// for message board
	delete dibMixed;
	dibMixed = NULL;
#endif
#ifdef USE_SMOOTH_STRETCH
	delete dibStretch1;
	dibStretch1 = NULL;
	delete dibStretch2;
	dibStretch2 = NULL;
#endif
#ifdef USE_DIRECT3D
	release_d3device();
	terminate_d3device();
#endif
	if (gui) {
		gui->ReleaseLedBox();
	}

	delete mux_update_screen;

	delete rec_video;
}

void EMU::release_screen_on_emu_thread()
{
	// stop video recording
	stop_rec_video();
}

void EMU::get_rgbformat(uint32_t *r_mask, uint32_t *g_mask, uint32_t *b_mask, uint8_t *r_shift, uint8_t *g_shift, uint8_t *b_shift)
{
	*r_mask = 0x00ff0000;
	*g_mask = 0x0000ff00;
	*b_mask = 0x000000ff;
	*r_shift = 16;
	*g_shift = 8;
	*b_shift = 0;
}

void EMU::get_rgbaformat(uint32_t *r_mask, uint32_t *g_mask, uint32_t *b_mask, uint32_t *a_mask, uint8_t *r_shift, uint8_t *g_shift, uint8_t *b_shift, uint8_t *a_shift)
{
	*a_mask = 0xff000000;
	*r_mask = 0x00ff0000;
	*g_mask = 0x0000ff00;
	*b_mask = 0x000000ff;
	*a_shift = 24;
	*r_shift = 16;
	*g_shift = 8;
	*b_shift = 0;
}

void EMU::get_rgbcolor(scrntype pixel, uint8_t *r, uint8_t *g, uint8_t *b)
{
	*r = (pixel & 0x00ff0000) >> 16;
	*g = (pixel & 0x0000ff00) >> 8;
	*b = (pixel & 0x000000ff);
}

scrntype EMU::map_rgbcolor(uint8_t r, uint8_t g, uint8_t b)
{
	return (scrntype)RGB_COLOR(r, g, b);
}

#if 0
int EMU::get_window_width(int mode)
{
#ifdef USE_SCREEN_ROTATE
	if(config.monitor_type) {
		return window_height + screen_height_aspect * mode;
	}
#endif
	return window_width + screen_width_aspect * mode;
}

int EMU::get_window_height(int mode) {
#ifdef USE_SCREEN_ROTATE
	if(config.monitor_type) {
		return window_width + screen_width_aspect * mode;
	}
#endif
	return window_height + screen_height_aspect * mode;
}
#endif

/// setting window or fullscreen size
/// @param [in] width
/// @param [in] height
/// @param [in] power magnify x 10
/// @param [in] window_mode  true:window / false:fullscreen
void EMU::set_display_size(int width, int height, int power, bool window_mode)
{
	lock_screen();

	bool display_size_changed = false;
	bool stretch_changed = false;

	if(width != -1 && (display_width != width || display_height != height)) {
		display_width = width;
		display_height = height;
		display_size_changed = stretch_changed = true;
	}

#ifdef USE_SCREEN_ROTATE
	int prev_source_width = source_width;
	int prev_source_height = source_height;

	if(config.monitor_type) {
		dibSource = dibRotate;

		stretch_changed |= (source_width != screen_height);
		stretch_changed |= (source_height != screen_width);
		stretch_changed |= (source_width_aspect != screen_height_aspect);
		stretch_changed |= (source_height_aspect != screen_width_aspect);

		source_width = screen_height;
		source_height = screen_width;
		source_width_aspect = screen_height_aspect;
		source_height_aspect = screen_width_aspect;
	}
	else {
#endif
		dibSource = dibOrigin;

		stretch_changed |= (source_width != screen_width);
		stretch_changed |= (source_height != screen_height);
		stretch_changed |= (source_width_aspect != screen_width_aspect);
		stretch_changed |= (source_height_aspect != screen_height_aspect);

		source_width = screen_width;
		source_height = screen_height;
		source_width_aspect = screen_width_aspect;
		source_height_aspect = screen_height_aspect;
#ifdef USE_SCREEN_ROTATE
	}
#ifdef USE_EMU_INHERENT_SPEC
	if (source_width != prev_source_width && source_height != prev_source_height) {
		HDC hdc = GetDC(main_window_handle);
		// for message board
		dibMixed->ReleaseDib();
		dibMixed->CreateDib(hdc, source_width, source_height);
		ReleaseDC(main_window_handle, hdc);
	}
#endif
#endif

	// fullscreen and stretch screen
	if(config.stretch_screen && !window_mode && display_width >= source_width && display_height >= source_height) {
		if (config.stretch_screen == 1) {
			// fit to full screen
			mixed_width = source_width;
			mixed_height = source_height;
			if (mixed_width_ratio < mixed_height_ratio) {
				mixed_height = mixed_height * mixed_width_ratio / mixed_height_ratio;
			} else {
				mixed_width = mixed_width * mixed_height_ratio / mixed_width_ratio;
			}
			mixed_dest_x = (source_width - mixed_width) / 2;
			mixed_dest_y = (source_height - mixed_height) / 2;
			mixed_dest_y = adjust_y_position(mixed_height, mixed_dest_y);

			stretched_width = (display_height * source_width_aspect) / source_height_aspect;
			stretched_height = display_height;
			stretched_dest_real_x = - mixed_dest_x * display_height / source_height_aspect;
			stretched_dest_real_y = - mixed_dest_y * display_height / source_height_aspect;
			if(stretched_width > display_width) {
				stretched_width = display_width;
				stretched_height = (display_width * source_height_aspect) / source_width_aspect;
				stretched_dest_real_x = - mixed_dest_x * display_width / source_width_aspect;
				stretched_dest_real_y = - mixed_dest_y * display_width / source_width_aspect;
			}
			if (mixed_width_ratio < mixed_height_ratio) {
				stretched_dest_real_y = stretched_dest_real_y * mixed_height_ratio / mixed_width_ratio;
			} else {
				stretched_dest_real_x = stretched_dest_real_x * mixed_width_ratio / mixed_height_ratio;
			}
			stretched_dest_x = (display_width - stretched_width) / 2;
			stretched_dest_y = (display_height - stretched_height) / 2;
			stretched_dest_real_x += stretched_dest_x;
			stretched_dest_real_y += stretched_dest_y;
		} else {
			// fit text area to full screen (cut off padding area)
			int sw;
			int sh;
			sw = LIMIT_MIN_WINDOW_WIDTH;
			sh = LIMIT_MIN_WINDOW_HEIGHT;
			int mw = sw;
			int mh = sh;
			if (mixed_width_ratio < mixed_height_ratio) {
				mh = mh * mixed_height_ratio / mixed_width_ratio;
			} else {
				mw = mw * mixed_width_ratio / mixed_height_ratio;
			}
			int mmw = 0;
			int mmh = 0;

			stretched_width = (display_height * mw) / mh;
			stretched_height = display_height;
			stretched_dest_real_x = 0;
			stretched_dest_real_y = 0;
			mmw = (display_width * mh / display_height) - mw;
			mmh = 0;
			if(stretched_width > display_width) {
				stretched_width = display_width;
				stretched_height = (display_width * mh) / mw;
				mmw = 0;
				mmh = (display_height * mw / display_width) - mh;

				stretched_height = display_height;

				if (sh + mmh > source_height) {
					mmh = sh + mmh - source_height;
					stretched_height -= mmh * display_width / mw;
					stretched_dest_real_y = 0;
				} else {
					stretched_dest_real_y = (sh + mmh - source_height) * display_width / mw / 2;
				}
				stretched_dest_real_x = (sw - source_width) * display_width / sw / 2;
			} else {
				stretched_width = display_width;

				if (sw + mmw > source_width) {
					mmw = sw + mmw - source_width;
					stretched_width -= mmw * display_height / mh;
					stretched_dest_real_x = 0;
				} else {
					stretched_dest_real_x = (sw + mmw - source_width) * display_height / mh / 2;
				}
				stretched_dest_real_y = (sh - source_height) * display_height / sh / 2;
			}
			mixed_width  = sw + mmw;
			mixed_height = sh + mmh;
			mixed_dest_x = (source_width - mixed_width) / 2;
			mixed_dest_y = (source_height - mixed_height) / 2;
			mixed_dest_y = adjust_y_position(mixed_height, mixed_dest_y);

			stretched_dest_x = (stretched_width  < display_width  ? (display_width  - stretched_width) / 2 : 0);
			stretched_dest_y = (stretched_height < display_height ? (display_height - stretched_height) / 2 : 0);
			stretched_dest_real_x += stretched_dest_x;
			stretched_dest_real_y += stretched_dest_y;
		}
	}
	// window or non-streach mode
	else {
		for(int n = 0; n <= 1; n++) {
			if (n == 0) {
				mixed_width = display_width * 10 / power;
				mixed_height = display_height * 10 / power;
			} else {
				mixed_width = source_width;
				mixed_height = source_height;
			}
			if (mixed_width_ratio < mixed_height_ratio) {
				mixed_height = mixed_height * mixed_width_ratio / mixed_height_ratio;
			} else {
				mixed_width = mixed_width * mixed_height_ratio / mixed_width_ratio;
			}
			mixed_dest_x = (source_width - mixed_width) / 2;
			mixed_dest_y = (source_height - mixed_height) / 2;
			if (mixed_dest_x >= 0 && mixed_dest_y >= 0) {
				break;
			}
		}
		mixed_dest_y = adjust_y_position(mixed_height, mixed_dest_y);

		stretched_width = source_width_aspect * power / 10;
		stretched_height = source_height_aspect * power / 10;
		stretched_dest_x = (display_width - stretched_width) / 2;
		stretched_dest_y = (display_height - stretched_height) / 2;
		stretched_dest_real_x = - mixed_dest_x * power / 10;
		stretched_dest_real_y = - mixed_dest_y * power / 10;
		if (mixed_width_ratio < mixed_height_ratio) {
			stretched_dest_real_y = stretched_dest_real_y * mixed_height_ratio / mixed_width_ratio;
		} else {
			stretched_dest_real_x = stretched_dest_real_x * mixed_width_ratio / mixed_height_ratio;
		}
		if (stretched_dest_x < 0) {
			stretched_dest_x = 0;
			stretched_width = display_width;
		}
		if (stretched_dest_y < 0) {
			stretched_dest_y = 0;
			stretched_height = display_height;
		}
		stretched_dest_real_x += stretched_dest_x;
		stretched_dest_real_y += stretched_dest_y;
	}

#ifdef USE_SMOOTH_STRETCH
	int new_pow_x = 1, new_pow_y = 1;
	while(stretched_width > source_width * new_pow_x) {
		new_pow_x++;
	}
	while(stretched_height > source_height * new_pow_y) {
		new_pow_y++;
	}

	// support high quality stretch only for x1 window size in gdi mode
	if(new_pow_x > 1 && new_pow_y > 1) {
		new_pow_x = new_pow_y = 1;
	}

//	if(stretched_width == source_width * new_pow_x && stretched_height == source_height * new_pow_y) {
//		new_pow_x = new_pow_y = 1;
//	}
	if(stretch_pow_x != new_pow_x || stretch_pow_y != new_pow_y) {
		stretch_pow_x = new_pow_x;
		stretch_pow_y = new_pow_y;
		stretch_changed = true;
	}

	if(stretch_changed) {
		dibStretch1->ReleaseDib();
		dibStretch2->ReleaseDib();
		stretch_screen = false;
		if(stretch_pow_x != 1 || stretch_pow_y != 1) {
			HDC hdc = GetDC(main_window_handle);
			dibStretch1->CreateDib(hdc, source_width * stretch_pow_x, source_height * stretch_pow_y);
			SetStretchBltMode(dibStretch1->GetDC(), COLORONCOLOR);

			dibStretch2->CreateDib(hdc, stretched_width, stretched_height);
			SetStretchBltMode(dibStretch2->GetDC(), COLORONCOLOR);

			ReleaseDC(main_window_handle, hdc);
			stretch_screen = true;
		}
	}
#endif

	change_rec_video_size(config.screen_video_size);

//	first_draw_screen = false;
	first_draw_screen = true;
	first_invalidate = true;
	screen_size_changed = false;
#ifdef _DEBUG_LOG
	out_debugf(_T("set_display_size: w:%d h:%d power:%d %s"),width,height,power,window_mode ? _T("window") : _T("fullscreen"));
	out_debugf(_T("         display: w:%d h:%d"),display_width, display_height);
	out_debugf(_T("          screen: w:%d h:%d"), screen_width, screen_height);
	out_debugf(_T("          source: w:%d h:%d"), source_width, source_height);
	out_debugf(_T("           mixed: w:%d h:%d"), mixed_width, mixed_height);
	out_debugf(_T("         stretch: w:%d h:%d"), stretched_width, stretched_height);
//	out_debugf(_T("     screen dest: x:%d y:%d"), screen_dest_x, screen_dest_y);
	out_debugf(_T("      mixed dest: x:%d y:%d"), mixed_dest_x, mixed_dest_y);
	out_debugf(_T("    stretch dest: x:%d y:%d"), stretched_dest_x, stretched_dest_y);
	out_debugf(_T(" stretch dest re: x:%d y:%d"), stretched_dest_real_x, stretched_dest_real_y);
#endif

	SetRect(&reD3Dmix, mixed_dest_x, mixed_dest_y, mixed_dest_x + mixed_width, mixed_dest_y + mixed_height);
	SetRect(&reD3Dsuf, stretched_dest_x, stretched_dest_y, stretched_dest_x + stretched_width, stretched_dest_y + stretched_height);
#ifdef _DEBUG_LOG
	out_debugf(_T("D3D          mix: l:%d t:%d r:%d b:%d w:%d h:%d"),reD3Dmix.left,reD3Dmix.top,reD3Dmix.right,reD3Dmix.bottom,mixed_width,mixed_height);
	out_debugf(_T("             suf: l:%d t:%d r:%d b:%d"),reD3Dsuf.left,reD3Dsuf.top,reD3Dsuf.right,reD3Dsuf.bottom);
#endif
	if (window_mode) {
		stretched_dest_x += display_margin.left;
		stretched_dest_y += display_margin.top;
		reD3Dsuf.left += display_margin.left;
		reD3Dsuf.top += display_margin.top;
		reD3Dsuf.right += display_margin.left;
		reD3Dsuf.bottom += display_margin.top;
#ifdef _DEBUG_LOG
		out_debugf(_T(" margin      suf: l:%d t:%d r:%d b:%d"),reD3Dsuf.left,reD3Dsuf.top,reD3Dsuf.right,reD3Dsuf.bottom);
#endif
	}

#ifdef USE_DIRECT3D
	if (pD3Device) {
		HRESULT hre;

		if (window_mode) {
			d3dpp.BackBufferWidth = 0;
			d3dpp.BackBufferHeight = 0;
			d3dpp.Windowed = TRUE;
		} else {
			d3dpp.BackBufferWidth = display_width;
			d3dpp.BackBufferHeight = display_height;
			d3dpp.Windowed = TRUE;
		}
		// save buffer to dib temporary
		if (config.use_direct3d) {
			copy_d3dsuf_dib(pD3Dorigin, dibOrigin->GetBuffer(), true);
		}
		// re create buffer
		hre = reset_d3device(main_window_handle);
		if (hre == D3D_OK) {
			pD3Device->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,0,0), 0.0, 0);
			// restore buffer from dib
			copy_d3dsuf_dib(pD3Dorigin, dibOrigin->GetBuffer(), false);
		}
		if (pD3Device != NULL) {
			LPDIRECT3DSURFACE9 suf = NULL;
			D3DSURFACE_DESC desc;
			pD3Device->GetBackBuffer(0,0,D3DBACKBUFFER_TYPE_MONO,&suf);
			if (suf != NULL) {
				suf->GetDesc(&desc);
				out_debugf(_T("d3dbackbuffer: w:%d h:%d"),desc.Width,desc.Height);
				suf->Release();

				if (desc.Width < (UINT)(reD3Dsuf.right - reD3Dsuf.left)) {
					// adjust width to fit the surface
					uint32_t new_mw = (uint32_t)((uint64_t)desc.Width * (reD3Dmix.right - reD3Dmix.left) / (reD3Dsuf.right - reD3Dsuf.left));
					uint32_t new_sw = (uint32_t)((uint64_t)new_mw * (reD3Dsuf.right - reD3Dsuf.left) / (reD3Dmix.right - reD3Dmix.left));
					reD3Dmix.right = reD3Dmix.left + new_mw;
					reD3Dsuf.right = reD3Dsuf.left + new_sw;
				}
				if (desc.Height < (UINT)(reD3Dsuf.bottom - reD3Dsuf.top)) {
					// adjust height to fit the surface
					uint32_t new_mh = (uint32_t)((uint64_t)desc.Height * (reD3Dmix.bottom - reD3Dmix.top) / (reD3Dsuf.bottom - reD3Dsuf.top));
					uint32_t new_sh = (uint32_t)((uint64_t)new_mh * (reD3Dsuf.bottom - reD3Dsuf.top) / (reD3Dmix.bottom - reD3Dmix.top));
					reD3Dmix.bottom = reD3Dmix.top + new_mh;
					reD3Dsuf.bottom = reD3Dsuf.top + new_sh;
				}

				out_debugf(_T("          mix: l:%d t:%d r:%d b:%d"),reD3Dmix.left,reD3Dmix.top,reD3Dmix.right,reD3Dmix.bottom);
				out_debugf(_T("          suf: l:%d t:%d r:%d b:%d"),reD3Dsuf.left,reD3Dsuf.top,reD3Dsuf.right,reD3Dsuf.bottom);
			}
		}
#ifdef USE_SCREEN_ROTATE
		if(config.monitor_type) {
			pD3Dsource = pD3Drotate;
		} else
#endif
		{
			pD3Dsource = pD3Dorigin;
		}
	}
#endif	/* USE_DIRECT3D */

	// send display size to vm
	set_vm_display_size();

	if (gui) {
		gui->SetLedBoxPosition(window_mode, mixed_dest_x, mixed_dest_y, mixed_width, mixed_height, config.led_pos);
	}
#ifdef USE_MESSAGE_BOARD
	if (msgboard) {
		msgboard->SetSize(source_width, source_height);
		msgboard->SetMessagePos(4 + mixed_dest_x,  - 4 - source_height + mixed_dest_y + mixed_height, 2);
		msgboard->SetInfoPos(-4 - mixed_dest_x, 4 + mixed_dest_y, 1);
	}
#endif

	unlock_screen();
}

#if 0
void EMU::change_screen_size(int sw, int sh, int swa, int sha, int ww, int wh)
{
	// virtual machine changes the screen size
	if(screen_width != sw || screen_height != sh) {
		screen_width = sw;
		screen_height = sh;
		screen_width_aspect = (swa != -1) ? swa : sw;
		screen_height_aspect = (sha != -1) ? sha : sh;
		window_width = ww;
		window_height = wh;
		screen_size_changed = true;

		// re-create dib sections
		HDC hdc = GetDC(main_window_handle);
		dibOrigin->ReleaseDib();
		dibOrigin->CreateDib(hdc, screen_width, screen_height);
#ifdef USE_SCREEN_ROTATE
		dibRotate->ReleaseDib();
		dibRotate->CreateDib(hdc, screen_height, screen_width);
#endif
#ifdef USE_EMU_INHERENT_SPEC
		// for message board
		dibMixed->ReleaseDib();
		dibMixed->CreateDib(hdc, screen_width, screen_height);
#endif
		ReleaseDC(main_window_handle, hdc);

		// stop recording
		if(now_rec_video()) {
			stop_rec_video();
			stop_rec_sound();
		}

		// change the window size
		PostMessage(main_window_handle, WM_RESIZE, 0L, 0L);
	}
}
#endif

/// set vm drawing area
void EMU::set_vm_display_size()
{
	if (vm) {
		int new_size = config.screen_video_size;
		int l = mixed_dest_x;
		int t = mixed_dest_y;
		int r = l + mixed_width;
		int b = t + mixed_height;

		// vm drawing area is same as area to capture and record screen
		if (rec_video_size[new_size].w > mixed_width) {
			l = rec_video_size[new_size].x;
			r = rec_video_size[new_size].x + rec_video_size[new_size].w;
		}

		if (rec_video_size[new_size].h > mixed_height) {
			t = rec_video_size[new_size].y;
			b = rec_video_size[new_size].y + rec_video_size[new_size].h;
		}

		vm->set_display_size(l,t,r,b);
	}
}

/// reserve top margin area for ledbox
int EMU::adjust_y_position(int h, int y)
{
	if (20 < h - LIMIT_MIN_WINDOW_HEIGHT && h - LIMIT_MIN_WINDOW_HEIGHT < 42) {
		y -= (42 - h + LIMIT_MIN_WINDOW_HEIGHT) / 2;
	}
	return y;
}

///
/// draw src screen from virtual machine
///
void EMU::draw_screen()
{
	// don't draw screen before new screen size is applied to buffers
	if(screen_size_changed) {
		return;
	}

	lock_screen();

#ifdef USE_DIRECT3D
	HRESULT hre;
	D3DLOCKED_RECT pLockedRect;

	lpD3DBmp = NULL;
	if (pD3Dorigin) {
		hre = pD3Dorigin->LockRect(&pLockedRect, NULL, 0);
		if (hre == D3D_OK) {
			lpD3DBmp = (scrntype *)pLockedRect.pBits;
		}
	}
#endif
	// draw screen
	if (!config.now_power_off) {
		vm->draw_screen();
	} else {
		fill_gray();
	}

#ifdef USE_DIRECT3D
	if (pD3Dorigin)	{
		pD3Dorigin->UnlockRect();
	}
#endif

	unlock_screen();

	// screen size was changed in vm->draw_screen()
	if(screen_size_changed) {
		return;
	}

#ifdef USE_SCREEN_ROTATE
	// rotate screen
	// right turn
	if(config.monitor_type) {
#ifdef USE_DIRECT3D
		if (config.use_direct3d) {
			D3DLOCKED_RECT pLockedRectOrigin;
			D3DLOCKED_RECT pLockedRectRotate;
			hre = pD3Dorigin->LockRect(&pLockedRectOrigin, NULL, 0);
			hre = pD3Drotate->LockRect(&pLockedRectRotate, NULL, 0);
			for(int y = 0; y < screen_height; y++) {
				scrntype* src = (scrntype *)pLockedRectOrigin.pBits + screen_width * y;
				scrntype* out = (scrntype *)pLockedRectRotate.pBits + (screen_height - y - 1);
				for(int x = 0; x < screen_width; x++) {
					*out = src[x];
					out += screen_height;
				}
			}
			pD3Drotate->UnlockRect();
			pD3Dorigin->UnlockRect();
		} else
#endif
		{
			for(int y = 0; y < screen_height; y++) {
				scrntype* src = dibOrigin->GetBuffer() + screen_width * (screen_height - y - 1);
				scrntype* out = dibRotate->GetBuffer() + screen_height * (screen_width - 1) + (screen_height - y - 1);
				for(int x = 0; x < screen_width; x++) {
					*out = src[x];
					out -= screen_height;
				}
			}
		}
	}
#endif

	lock_screen();

#ifdef USE_DIRECT3D
	if (config.use_direct3d) {
		if (gui) {
			gui->DrawLedBox(pD3Dsource);
		}
	} else
#endif
	{
		if (gui) {
			gui->DrawLedBox(dibSource->GetDC());
		}
	}

	unlock_screen();
}

void EMU::fill_gray()
{
	scrntype color = map_rgbcolor(0x40, 0x40, 0x40);
	for(int y=0; y<source_height; y++) {
		scrntype *line = screen_buffer(y);
		for(int x=0; x<source_width; x++) {
			*line = color;
			line++;
		}
	}
}

///
/// copy src screen to mix screen and overlap a message
///
void EMU::mix_screen()
{
#ifdef USE_EMU_INHERENT_SPEC
#ifdef USE_DIRECT3D
	if (pD3Device != NULL && config.use_direct3d) {
		pD3Device->UpdateSurface(pD3Dsource, NULL, pD3Dmixsuf, NULL);
#ifdef USE_MESSAGE_BOARD
		if (msgboard) {
			msgboard->Draw(pD3Dmixsuf);
		}
#endif
	} else
#endif
	{
		BitBlt(dibMixed->GetDC(), 0, 0, source_width, source_height, dibSource->GetDC(), 0, 0, SRCCOPY);
#ifdef USE_MESSAGE_BOARD
		if (msgboard) {
			msgboard->Draw(dibMixed->GetDC());
		}
#endif
	}
#endif

	// stretch screen
#ifdef USE_SMOOTH_STRETCH
	//	if(stretch_screen) {
	if(stretch_screen && (!config.use_direct3d)) {
#if 0
		StretchBlt(dibStretch1->GetDC(), 0, 0, source_width * stretch_pow_x, source_height * stretch_pow_y, dibSource->GetDC(), 0, 0, source_width, source_height, SRCCOPY);
#else
		// about 50% faster than StretchBlt()
		scrntype* src = dibSource->GetBuffer() + source_width * (source_height - 1);
		scrntype* out = dibStretch1->GetBuffer() + source_width * stretch_pow_x * (source_height * stretch_pow_y - 1);
		int data_len = source_width * stretch_pow_x;

		for(int y = 0; y < source_height; y++) {
			if(stretch_pow_x != 1) {
				scrntype* out_tmp = out;
				for(int x = 0; x < source_width; x++) {
					scrntype c = src[x];
					for(int px = 0; px < stretch_pow_x; px++) {
						out_tmp[px] = c;
					}
					out_tmp += stretch_pow_x;
				}
			}
			else {
				// faster than memcpy()
				for(int x = 0; x < source_width; x++) {
					out[x] = src[x];
				}
			}
			if(stretch_pow_y != 1) {
				scrntype* src_tmp = out;
				for(int py = 1; py < stretch_pow_y; py++) {
					out -= data_len;
					// about 10% faster than memcpy()
					for(int x = 0; x < data_len; x++) {
						out[x] = src_tmp[x];
					}
				}
			}
			src -= source_width;
			out -= data_len;
		}
#endif

		StretchBlt(dibStretch2->GetDC(), 0, 0, stretched_width, stretched_height, dibStretch1->GetDC(), 0, 0, source_width * stretch_pow_x, source_height * stretch_pow_y, SRCCOPY);
	}
#endif
	first_draw_screen = true;
}

///
/// post request screen updating to draw it on main thread
///
void EMU::need_update_screen()
{
	// invalidate window
	UINT flags = RDW_INVALIDATE | RDW_INTERNALPAINT | (first_invalidate ? RDW_ERASE : RDW_NOERASE);
	RedrawWindow(main_window_handle, NULL, NULL, flags);
	self_invalidate = true;
//	skip_frame = false;
}

///
/// return pointer on source screen
///
scrntype* EMU::screen_buffer(int y)
{
#ifdef USE_DIRECT3D
	if (lpD3DBmp && config.use_direct3d) {
		return lpD3DBmp + screen_width * y;
	} else
#endif
	return dibOrigin->GetBuffer() + screen_width * (screen_height - y - 1);
}

int EMU::screen_buffer_offset()
{
#ifdef USE_DIRECT3D
	if (lpD3DBmp && config.use_direct3d) {
		return screen_width;
	} else
#endif
	return -screen_width;
}

#ifdef USE_DIRECT3D
void EMU::update_screen_d3d()
{
	HRESULT hre = S_OK;
	PDIRECT3DSURFACE9 pD3Dsuf = NULL;

	if (lost_d3device) {
		// can use a device?
		hre = pD3Device->TestCooperativeLevel();
		if (hre == D3DERR_DEVICELOST) {
			// now lost device
			lost_d3device = true;
		} else if (hre == D3DERR_DEVICENOTRESET) {
			// reset device to use
			if (reset_d3device(main_window_handle) == D3D_OK) {
				lost_d3device = false;
			}
		} else {
			lost_d3device = false;
		}
	} else {
		if(first_draw_screen && pD3Device != NULL) {
			// get back buffer
			pD3Device->GetBackBuffer(0,0,D3DBACKBUFFER_TYPE_MONO,&pD3Dsuf);
			if (pD3Dsuf) {
				// copy to back buffer from offline surface
				hre = pD3Device->StretchRect(pD3Dmixsuf, &reD3Dmix, pD3Dsuf, &reD3Dsuf, (_D3DTEXTUREFILTERTYPE)config.d3d_filter_type);

				pD3Dsuf->Release();
			}
			// draw
			hre = pD3Device->Present(NULL,NULL,NULL,NULL);

			first_invalidate = self_invalidate = false;
		}
		if (hre == D3DERR_DEVICELOST) {
			// now lost device
			lost_d3device = true;
		}
	}
}
#endif

///
/// render screen and display window
/// @note must be called by main thread
void EMU::update_screen(HDC hdc)
{
	if (!initialized) return;

	lock_screen();

	mix_screen();

#ifdef USE_DIRECT3D
	if (pD3Device != NULL && config.use_direct3d) {
		update_screen_d3d();
	} else
#endif
	{
		update_screen_dc(hdc);
	}

	if (gui) {
		gui->UpdateIndicator(update_led());
	}

	unlock_screen();
}

void EMU::update_screen_dc(HDC hdc)
{
#ifdef USE_BITMAP
	if(first_invalidate || !self_invalidate) {
		HDC hmdc = CreateCompatibleDC(hdc);
		HBITMAP hBitmap = LoadBitmap(instance_handle, _T("IDI_BITMAP1"));
		BITMAP bmp;
		GetObject(hBitmap, sizeof(BITMAP), &bmp);
		int w = (int)bmp.bmWidth;
		int h = (int)bmp.bmHeight;
		HBITMAP hOldBitmap = (HBITMAP)SelectObject(hmdc, hBitmap);
		BitBlt(hdc, 0, 0, w, h, hmdc, 0, 0, SRCCOPY);
		SelectObject(hmdc, hOldBitmap);
		DeleteObject(hBitmap);
		DeleteDC(hmdc);
	}
#endif
	if(first_draw_screen) {
#ifdef USE_LED
		// 7-seg LEDs
		for(int i = 0; i < MAX_LEDS; i++) {
			int x = leds[i].x;
			int y = leds[i].y;
			int w = leds[i].width;
			int h = leds[i].height;
			BitBlt(hdc, x, y, w, h, hdcDib, x, y, SRCCOPY);
		}
#else
		// standard screen
#ifdef USE_SMOOTH_STRETCH
		if(stretch_screen) {
			BitBlt(hdc, screen_dest_x, screen_dest_y, stretched_width, stretched_height, dibStretch2->GetDC(), 0, 0, SRCCOPY);
		}
		else {
#endif
#ifdef USE_EMU_INHERENT_SPEC
			if(stretched_width == mixed_width && stretched_height == mixed_height) {
				BitBlt(hdc, stretched_dest_x, stretched_dest_y, stretched_width, stretched_height, dibMixed->GetDC(), reD3Dmix.left, reD3Dmix.top, SRCCOPY);
			}
			else {
				StretchBlt(hdc, stretched_dest_x, stretched_dest_y, stretched_width, stretched_height, dibMixed->GetDC(), reD3Dmix.left, reD3Dmix.top, mixed_width, mixed_height, SRCCOPY);
			}
#else
			if(stretched_width == source_width && stretched_height == source_height) {
				BitBlt(hdc, screen_dest_x, screen_dest_y, stretched_width, stretched_height, hdcDibSource, 0, 0, SRCCOPY);
			}
			else {
				StretchBlt(hdc, screen_dest_x, screen_dest_y, stretched_width, stretched_height, hdcDibSource, 0, 0, source_width, source_height, SRCCOPY);
			}
#endif
#ifdef USE_SMOOTH_STRETCH
		}
#endif

#ifdef USE_ACCESS_LAMP
		// draw access lamps of drives
		int status = vm->access_lamp() & 7;
		static int prev_status = 0;
		bool render_in = (status != 0);
		bool render_out = (prev_status != status);
		prev_status = status;

		if(render_in || render_out) {
			COLORREF crColor = RGB((status & 1) ? 255 : 0, (status & 2) ? 255 : 0, (status & 4) ? 255 : 0);
			int right_bottom_x = screen_dest_x + stretched_width;
			int right_bottom_y = screen_dest_y + stretched_height;

			for(int y = display_height - 6; y < display_height; y++) {
				for(int x = display_width - 6; x < display_width; x++) {
					if((x < right_bottom_x && y < right_bottom_y) ? render_in : render_out) {
						SetPixelV(hdc, x, y, crColor);
					}
				}
			}
		}
#endif
#endif
		first_invalidate = self_invalidate = false;
	}
}

#ifdef USE_DIRECT3D
void EMU::copy_d3dsuf_dib(PDIRECT3DSURFACE9 suf, scrntype *buf, bool to_dib)
{
	HRESULT hre;
	D3DLOCKED_RECT pLockedRect;

	hre = suf->LockRect(&pLockedRect, NULL, 0);
	if (hre == D3D_OK) {
		scrntype *src;
		scrntype *out;
		if (to_dib) {
			src = (scrntype *)pLockedRect.pBits;
			out = buf;
		} else {
			src = buf;
			out = (scrntype *)pLockedRect.pBits;
		}
		out += screen_width * (screen_height - 1);
		int data_len = screen_width;

		for(int y = 0; y < screen_height; y++) {
			for(int i = 0; i < data_len; i++) {
				out[i] = src[i];
			}
			src += data_len;
			out -= data_len;
		}
		suf->UnlockRect();
	}
}
#endif

bool EMU::now_skip_frame()
{
	return skip_frame && !now_rec_video();
}

///
/// capture current screen and save to a file
///
void EMU::capture_screen()
{
	lock_screen();

#ifdef USE_DIRECT3D
	if (pD3Dorigin != NULL && config.use_direct3d) {
		copy_d3dsuf_dib(pD3Dorigin, dibOrigin->GetBuffer(), true);
	}
#endif
//	copy_dib_rec_video();

	rec_video->Capture(CAPTURE_SCREEN_TYPE, &rec_video_stretched_size, dibOrigin, &rec_video_size[config.screen_video_size]);

	unlock_screen();
}

///
/// start recording video
///
bool EMU::start_rec_video(int type, int fps, bool show_dialog)
{
#ifdef USE_REC_VIDEO
	int size = config.screen_video_size;
	return rec_video->Start(type, fps, &rec_video_size[size], dibOrigin, show_dialog);
#else
	return false;
#endif
}

///
/// stop recording video
///
void EMU::stop_rec_video()
{
#ifdef USE_REC_VIDEO
	rec_video->Stop();
#endif
}

///
/// restart recording video
///
void EMU::restart_rec_video()
{
#ifdef USE_REC_VIDEO
	rec_video->Restart();
#endif
}

///
/// record video
///
void EMU::record_rec_video()
{
#ifdef USE_REC_VIDEO
	if (rec_video->IsRecordFrame()) {
#ifdef USE_DIRECT3D
		if (pD3Device != NULL && config.use_direct3d) {
			copy_d3dsuf_dib(pD3Dorigin, dibOrigin->GetBuffer(), true);
		}
#endif
		rec_video->Record(&rec_video_stretched_size, dibOrigin, &rec_video_size[config.screen_video_size]);
	}
#endif
}

bool EMU::now_rec_video()
{
#ifdef USE_REC_VIDEO
	return *now_recording_video;
#else
	return false;
#endif
}

void EMU::resize_rec_video(int num)
{
#ifdef USE_EMU_INHERENT_SPEC
	config.screen_video_size = num;
	change_rec_video_size(num);

	// send display size to vm
	set_vm_display_size();
#endif
}
void EMU::change_rec_video_size(int num)
{
	rec_video_stretched_size = rec_video_size[num];
	if (mixed_width_ratio < mixed_height_ratio) {
		rec_video_stretched_size.h = rec_video_stretched_size.h * mixed_width_ratio / mixed_height_ratio;
		rec_video_stretched_size.y += (rec_video_size[num].h - rec_video_stretched_size.h) / 2; 
	} else {
		rec_video_stretched_size.w = rec_video_stretched_size.w * mixed_height_ratio / mixed_width_ratio;
		rec_video_stretched_size.x += (rec_video_size[num].w - rec_video_stretched_size.w) / 2; 
	}
	rec_video_stretched_size.y = adjust_y_position(rec_video_stretched_size.h, rec_video_stretched_size.y);

#ifdef _DEBUG_LOG
	out_debugf(_T("          rec_video_size: num:%d x:%d y:%d w:%d h:%d")
		,num,rec_video_size[num].x,rec_video_size[num].y,rec_video_size[num].w,rec_video_size[num].h);
	out_debugf(_T("rec_video_stretched_size:       x:%d y:%d w:%d h:%d")
		,rec_video_stretched_size.x,rec_video_stretched_size.y,rec_video_stretched_size.w,rec_video_stretched_size.h);
#endif
}

bool EMU::rec_video_enabled(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->IsEnabled(type);
#else
	return false;
#endif
}

const _TCHAR **EMU::get_rec_video_codec_list(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->GetCodecList(type);
#else
	return NULL;
#endif
}

const _TCHAR **EMU::get_rec_video_quality_list(int type)
{
#ifdef USE_REC_VIDEO
	return rec_video->GetQualityList(type);
#else
	return NULL;
#endif
}

//----------

int EMU::qsort_by_w_h_n(const void *a, const void *b)
{
	struct window_mode_st *a1 = (struct window_mode_st *)a;
	struct window_mode_st *b1 = (struct window_mode_st *)b;

	int rc = 0;
	if (!rc) rc = (a1->width - b1->width);
	if (!rc) rc = (a1->height - b1->height);
	if (!rc) rc = (a1->power - b1->power);
	return rc;
}

int EMU::qsort_by_n_w_h(const void *a, const void *b)
{
	struct screen_mode_st *a1 = (struct screen_mode_st *)a;
	struct screen_mode_st *b1 = (struct screen_mode_st *)b;

	int rc = 0;
	int aa = a1->width * a1->height;
	int bb = b1->width * b1->height;
	if (!rc) rc = (a1->disp_no - b1->disp_no);
	if (!rc) rc = (bb - aa);
	return rc;
}

BOOL CALLBACK EMU::MonitorProc(HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor, LPARAM dwData)
{
	EMU *emu = (EMU *)dwData;
	return emu->get_monitor_device(hMonitor, hdcMonitor, lprcMonitor);
}

BOOL EMU::get_monitor_device(HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor)
{
	if (monitor_count < disp_device_count) {
			disp_device[monitor_count].hMoni = hMonitor;
			CopyRect(&disp_device[monitor_count].re, lprcMonitor);
	}
	monitor_count++;

	return TRUE;
}

/// enumerate screen mode for window
/// calcurate magnify range
/// @param [in] max_width
/// @param [in] max_height
void EMU::enum_window_mode(int max_width, int max_height)
{
	// enumerate screen mode for window
	int w,h;
	bool minsize = false;
	bool maxsize = false;

	window_mode_count = 0;
	for (int i=0; window_proposal[i].cx != 0 && window_mode_count < WINDOW_MODE_MAX; i++) {
		w = window_proposal[i].cx;
		h = window_proposal[i].cy;
		// minimum window size?
		if (w < MIN_WINDOW_WIDTH || h < MIN_WINDOW_HEIGHT) {
			continue;
		}
		if (!minsize) {
			w = MIN_WINDOW_WIDTH;
			h = MIN_WINDOW_HEIGHT;
			i--;
			minsize = true;
		}
		if (w > MAX_WINDOW_WIDTH || h > MAX_WINDOW_HEIGHT) {
			if (!maxsize) {
				w = MAX_WINDOW_WIDTH;
				h = MAX_WINDOW_HEIGHT;
				maxsize = true;
				i--;
			} else {
				break;
			}
		}
		for (int pow = 10; pow <= 30 && window_mode_count < WINDOW_MODE_MAX; pow+=5) {
			if ((w*pow/10) <= max_width && (h*pow/10) <= max_height) {
				bool found = false;
				for(int j = 0; j < window_mode_count; j++) {
					if(window_mode[j].power == pow && window_mode[j].width == (w*pow/10) && window_mode[j].height == (h*pow/10)) {
						found = true;
						break;
					}
				}
				if (!found) {
					window_mode[window_mode_count].power = pow;
					window_mode[window_mode_count].width = w * pow / 10;
					window_mode[window_mode_count].height = h * pow / 10;
					out_debugf(_T("window_mode:%d %dx%d x%.1f"),window_mode_count, w * pow / 10, h * pow / 10, (double)pow / 10.0);
					window_mode_count++;
				}
			} else {
				break;
			}
		}
	}
	qsort(&window_mode, window_mode_count, sizeof(struct window_mode_st), qsort_by_w_h_n);
}

void EMU::enum_screen_mode()
{
	// enumerate display device
	int max_desktop_width = 0;
	int max_desktop_height = 0;
	disp_device_count = 0;
	for(int i = 0; disp_device_count < DISP_DEVICE_MAX; i++) {
		DISPLAY_DEVICE disp;
		ZeroMemory(&disp, sizeof(disp));
		disp.cb = sizeof(disp);
		if (!EnumDisplayDevices(NULL, i, &disp, 0)) {
			break;
		}
		if (disp.StateFlags & (DISPLAY_DEVICE_ACTIVE | DISPLAY_DEVICE_PRIMARY_DEVICE)) {
			_tcsncpy(disp_device[disp_device_count].name, disp.DeviceName, 32);
			disp_device[disp_device_count].hMoni = NULL;
			disp_device_count++;
		}
	}

	// enumerate screen mode for fullscreen
	int w,h,bpp;
	screen_mode_count = 0;
	for(int disp_no = 0; disp_no < disp_device_count && screen_mode_count < SCREEN_MODE_MAX; disp_no++) {
		for(int i = 0; screen_mode_count < SCREEN_MODE_MAX; i++) {
			DEVMODE dev;
			ZeroMemory(&dev, sizeof(dev));
			dev.dmSize = sizeof(dev);
			if(EnumDisplaySettings(disp_device[disp_no].name, i, &dev) == 0) {
				break;
			}
			w = dev.dmPelsWidth;
			h = dev.dmPelsHeight;
			bpp = dev.dmBitsPerPel;
			if(w < MIN_WINDOW_WIDTH || h < MIN_WINDOW_HEIGHT || w < 640 || h < 480) {
				out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp ignored"), disp_no, w, h, bpp);
				continue;
			}
			if(((w * 30 / h) != 40)
			&& ((w * 40 / h) != 50)
			&& ((w * 90 / h) != 160)
			&& ((w * 50 / h) != 80)) {
				out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp ignored"), disp_no, w, h, bpp);
				continue;
			}
			if (w <= 1280 && (w != 640 && w != 800 && w != 1024 && w != 1280)) {
				out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp ignored"), disp_no, w, h, bpp);
				continue;
			}
			bool found = false;
			for(int j = 0; j < screen_mode_count; j++) {
				if(screen_mode[j].disp_no == disp_no && screen_mode[j].width == w && screen_mode[j].height == h) {
					found = true;
					out_debugf(_T("screen_mode:-- [%d] %dx%d %dbpp already exist"), disp_no, w, h, bpp);
					break;
				}
			}
			if(!found) {
				screen_mode[screen_mode_count].disp_no = disp_no;
				screen_mode[screen_mode_count].width = w;
				screen_mode[screen_mode_count].height = h;
				out_debugf(_T("screen_mode:%2d [%d] %dx%d %dbpp"), screen_mode_count, disp_no, w, h, bpp);
				screen_mode_count++;

				if(max_desktop_width <= w) {
					max_desktop_width = w;
				}
				if(max_desktop_height <= h) {
					max_desktop_height = h;
				}
			}
		}
	}
	qsort(&screen_mode, screen_mode_count, sizeof(struct screen_mode_st), qsort_by_n_w_h);

	// if cannot get modes, add default screen size mode.
	if (screen_mode_count == 0) {
		screen_mode[screen_mode_count].disp_no = 0;
		screen_mode[screen_mode_count].width = desktop_width;
		screen_mode[screen_mode_count].height = desktop_height;
		screen_mode_count = 1;

		max_desktop_width = desktop_width;
		max_desktop_height = desktop_height;
	}

	// enumerate monitor size
	EnumDisplayMonitors(NULL, NULL, MonitorProc, (LPARAM)this);
}

/// matching window or fullscreen size
void EMU::find_screen_mode()
{
	if (config.window_mode < 8) {
		// matching window size at last time
		bool find = false;
		int i = config.window_mode;
		if (i < window_mode_count) {
			if (window_mode[i].width == config.screen_width
			&&  window_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	} else {
		// matching fullscreen size at last time
		bool find = false;
		for(int i = 0; i < screen_mode_count; i++) {
			if (screen_mode[i].disp_no == config.disp_device_no
			&&  screen_mode[i].width == config.screen_width
			&&  screen_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i + 8;
				break;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	}
}

/// change window size / switch over fullscreen and window
/// @param[in] mode 0 - 7: window size  8 -:  fullscreen size  -1: switch over  -2: shift window mode
void EMU::change_screen_mode(int mode)
{
//	out_debugf(_T("change_screen_mode: mode:%d cwmode:%d pwmode:%d w:%d h:%d"),mode,config.window_mode,prev_window_mode,desktop_width,desktop_height);
//	if (mode == config.window_mode) return;

	if (mode == -1) {
		// switch over fullscreen and window
		if (now_fullscreen) {
			// go window mode
			mode = prev_window_mode;
		}
	} else if (mode == -2) {
		// shift window mode 
		if (now_fullscreen) {
			// no change
			return;
		} else {
			mode = ((config.window_mode + 1) % window_mode_count);
		}
	}
	if (!now_fullscreen) {
		prev_window_mode = config.window_mode;
	}
//	out_debugf(_T("change_screen_mode: mode:%d cwmode:%d pwmode:%d w:%d h:%d"),mode,config.window_mode,prev_window_mode,desktop_width,desktop_height);
	set_window(mode, desktop_width, desktop_height);
}

void EMU::change_stretch_screen(int num)
{
	if (num < 0) {
		num = (config.stretch_screen + 1);
	}
	num = (num % 3);
	if (config.stretch_screen == num) {
		num = 0;
	}
	config.stretch_screen = num;
	if (now_fullscreen) {
		set_display_size(-1, -1, 10, !now_fullscreen);
	}
}

#if 0
void EMU::change_cutout_screen()
{
	config.cutout_screen = !config.cutout_screen;
	if (now_fullscreen) set_display_size(-1, -1, 10, !now_fullscreen);
}
#endif

void EMU::change_pixel_aspect(int mode)
{
	if (mode < 0) {
		mode = (config.pixel_aspect + 1) % get_pixel_aspect_count();
	}
	config.pixel_aspect = get_pixel_aspect(mode, &mixed_width_ratio, &mixed_height_ratio);

	set_display_size(config.screen_width, config.screen_height, now_fullscreen ? 10 : window_mode_power, !now_fullscreen);
}

int EMU::get_pixel_aspect_count()
{
	return 3;
}

int EMU::get_pixel_aspect(int mode, int *wratio, int *hratio)
{
	int new_mode = mode;
	switch(mode) {
	case 1:
		*wratio = 10;
		*hratio = 11;
		break;
	case 2:
		*wratio = 10;
		*hratio = 12;
		break;
	default:
		*wratio = 1;
		*hratio = 1;
		new_mode = 0;
		break;
	}
	return new_mode;
}

void EMU::set_client_pos(int dest_x, int dest_y, int client_width, int client_height, UINT flags)
{
	WINDOWINFO wid;
	GetWindowInfo(main_window_handle, &wid);

	int width = (wid.rcClient.left - wid.rcWindow.left) + client_width + (wid.rcWindow.right - wid.rcClient.right);
	int height = (wid.rcClient.top - wid.rcWindow.top) + client_height + (wid.rcWindow.bottom - wid.rcClient.bottom);

	if (flags & 1) {
		// center position
		dest_x = (int)((desktop_width - width) / 2);
		dest_y = (int)((desktop_height - height) / 4);
//		dest_x = (dest_x < 0) ? 0 : dest_x;
		dest_y = (dest_y < 0) ? 0 : dest_y;
		flags &= ~1;
	} else {
		flags = SWP_NOMOVE;
	}
	flags |= SWP_NOZORDER;
#ifdef _DEBUG
	out_debugf(_T("set_client_pos: x:%d y:%d w:%d h:%d flags:%x"), dest_x, dest_y, width, height, flags);
#endif
	SetWindowPos(main_window_handle, HWND_TOP, dest_x, dest_y, width, height, flags);
}

/// setting window or fullscreen size to display
/// @param [in] mode 0 .. 7 window mode / 8 .. 23 fullscreen mode / -1 want to go fullscreen, but unknown mode
/// @param [in] cur_width  current desktop width if mode is -1
/// @param [in] cur_height current desktop height if mode is -1
void EMU::set_window(int mode, int cur_width, int cur_height)
{
	static LONG style = WS_VISIBLE;
	WINDOWPLACEMENT place;
	place.length = sizeof(WINDOWPLACEMENT);

	if(mode >= 0 && mode < 8) {
		// go window
		if (mode >= window_mode_count) mode = 0;
//		int width = emu->get_window_width(mode);
//		int height = emu->get_window_height(mode);
		int width = window_mode[mode].width;
		int height = window_mode[mode].height;
#ifdef USE_SCREEN_ROTATE
		if (config.monitor_type) {
			int v = width;
			width = height;
			height = v;
		}
#endif
//		RECT rect = {0, 0, width, height};
		RECT rect = {-display_margin.left, -display_margin.top, width + display_margin.right, height + display_margin.bottom};
//		AdjustWindowRectEx(&rect, dwStyle, TRUE, 0);
		int dest_x = 0;
		int dest_y = 0;
		if (first_change_screen) {
			// center position
			dest_x = (int)((desktop_width - (rect.right - rect.left)) / 2);
			dest_y = (int)((desktop_height - (rect.bottom - rect.top)) / 4);
//			dest_x = (dest_x < 0) ? 0 : dest_x;
			dest_y = (dest_y < 0) ? 0 : dest_y;
		}

		if(now_fullscreen) {
			// change fullscreen to window
			ChangeDisplaySettings(NULL, 0);
			SetWindowLong(main_window_handle, GWL_STYLE, style);
			if (!first_change_screen) {
				// rollback position of window mode
				dest_x = (int)window_dest.x;
				dest_y = (int)window_dest.y;
			}
#ifdef _DEBUG
			out_debugf(_T("set_window: f->w x:%d y:%d w:%d h:%d"), dest_x, dest_y, rect.right - rect.left, rect.bottom - rect.top);
#endif
			SetWindowPos(main_window_handle, HWND_TOP, dest_x, dest_y, rect.right - rect.left, rect.bottom - rect.top, SWP_SHOWWINDOW);
			now_fullscreen = false;
			// show menu
			gui->ScreenModeChanged(now_fullscreen);
			// resize client
			set_client_pos(dest_x, dest_y, rect.right - rect.left, rect.bottom - rect.top, first_change_screen ? 1 : 0);
		}
		else {
			// change window and client size
			if (!first_change_screen) {
				// keep current position if possible
				WINDOWINFO wid;
				GetWindowInfo(main_window_handle, &wid);
				dest_x = wid.rcWindow.left;
				dest_y = wid.rcWindow.top;
				if (dest_y + rect.bottom - rect.top > desktop_height) dest_y = desktop_height - rect.bottom + rect.top;
				if (dest_y < 0) dest_y = 0;
			}
#ifdef _DEBUG
			out_debugf(_T("set_window: w->w x:%d y:%d w:%d h:%d"), dest_x, dest_y, rect.right - rect.left, rect.bottom - rect.top);
#endif
			SetWindowPos(main_window_handle, NULL, dest_x, dest_y, rect.right - rect.left, rect.bottom - rect.top, SWP_NOZORDER);
			// resize client
			set_client_pos(dest_x, dest_y, rect.right - rect.left, rect.bottom - rect.top, first_change_screen ? 1 : 0);
		}
		config.window_mode = prev_window_mode = mode;
		config.disp_device_no = 0;
		config.screen_width = width;
		config.screen_height = height;
		window_mode_power = window_mode[mode].power;

		// set screen size to emu class
		set_display_size(width, height, window_mode_power, true);
#ifdef USE_MOUSE_ABSOLUTE
		set_mouse_position();
#endif
	}
	else if(!now_fullscreen) {
		// go fullscreen
		if (mode >= screen_mode_count + 8) mode = -1;
		int disp_no = (mode == -1) ? 0 : screen_mode[mode - 8].disp_no;
		int width = (mode == -1) ? cur_width : screen_mode[mode - 8].width;
		int height = (mode == -1) ? cur_height : screen_mode[mode - 8].height;
		int top = 0;
		int left = 0;

		// store current position of now window mode
		WINDOWINFO wid;
		GetWindowInfo(main_window_handle, &wid);
		window_dest.x = wid.rcWindow.left;
		window_dest.y = wid.rcWindow.top;

		DEVMODE dev;
		ZeroMemory(&dev, sizeof(dev));
		dev.dmSize = sizeof(dev);
		dev.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		dev.dmBitsPerPel = desktop_bpp;
		dev.dmPelsWidth = width;
		dev.dmPelsHeight = height;

		TCHAR *dev_name = NULL;
		if (disp_no < disp_device_count) {
			dev_name = disp_device[disp_no].name;
			top =  disp_device[disp_no].re.top;
			left = disp_device[disp_no].re.left;
		}
		// try fullscreen
		if(ChangeDisplaySettingsEx(dev_name, &dev, NULL, CDS_TEST, NULL) == DISP_CHANGE_SUCCESSFUL) {
			GetWindowPlacement(main_window_handle, &place);
			ChangeDisplaySettingsEx(dev_name, &dev, NULL, CDS_FULLSCREEN, NULL);
			style = GetWindowLong(main_window_handle, GWL_STYLE);
			SetWindowLong(main_window_handle, GWL_STYLE, WS_VISIBLE);
			SetWindowPos(main_window_handle, HWND_TOP, left, top, width, height, SWP_SHOWWINDOW);
			SetCursorPos(width / 2, height / 2);
			now_fullscreen = true;

			if(mode == -1) {
				for(int i = 0; i < screen_mode_count; i++) {
					if(screen_mode[i].disp_no == disp_no && screen_mode[i].width == width && screen_mode[i].height == height) {
						mode = i + 8;
						break;
					}
				}
			}
			config.window_mode = mode;
			config.disp_device_no = disp_no;
			config.screen_width = width;
			config.screen_height = height;

			// remove menu
			gui->ScreenModeChanged(true);

			// set screen size to emu class
			set_display_size(width, height, 10, false);

#ifdef USE_MOUSE_ABSOLUTE
			set_mouse_position();
#endif
		}
	} else {
		// now fullscreen
		if (mode >= screen_mode_count + 8) mode = -1;
		int width = (mode == -1) ? cur_width : screen_mode[mode - 8].width;
		int height = (mode == -1) ? cur_height : screen_mode[mode - 8].height;

		// set screen size to emu class
		set_display_size(width, height, 10, false);
	}
	first_change_screen = false;
//	mute_sound(false);
	set_pause(1, false);
}

window_mode_t *EMU::get_window_mode(int num)
{
	if (num < window_mode_count) {
		return &window_mode[num];
	}
	return NULL;
}

screen_mode_t *EMU::get_screen_mode(int num)
{
	if (num < screen_mode_count) {
		return &screen_mode[num];
	}
	return NULL;
}

VmRectWH *EMU::get_screen_record_size(int num)
{
	if (0 <= num && num <= 1) {
		return &rec_video_size[num];
	}
	return NULL;
}

// ----------
#ifdef USE_DIRECT3D
void EMU::initialize_d3device(HWND hWnd)
{
	ZeroMemory(&d3dpp , sizeof(D3DPRESENT_PARAMETERS));

	// D3D
	pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if (pD3D) {
		pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT , &d3ddm);

		d3dpp.BackBufferWidth = 0;
		d3dpp.BackBufferHeight = 0;
		d3dpp.BackBufferFormat = d3ddm.Format;
		d3dpp.BackBufferCount = 1;
		d3dpp.MultiSampleType = D3DMULTISAMPLE_NONE;
		d3dpp.MultiSampleQuality = 0;
		d3dpp.EnableAutoDepthStencil = FALSE;
		d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		d3dpp.Windowed = TRUE;

		set_d3dpresent_interval();
	}
}

HRESULT EMU::create_d3device(HWND hWnd)
{
	D3DCAPS9 d3dcaps;
	HRESULT hre;

	lost_d3device = false;

	hre = pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING ,&d3dpp, &pD3Device);
	if (hre != D3D_OK) {
		hre = pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING ,&d3dpp, &pD3Device);
	}
	if (hre != D3D_OK) {
		// disable direct 3d
		pD3Device = NULL;
	}
	enable_direct3d = false;

	if (pD3Device != NULL) {
		enable_direct3d = true;

		pD3Device->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,0,0), 0.0, 0);

		pD3Device->GetDeviceCaps(&d3dcaps);
		if ((d3dcaps.Caps & D3DCAPS_READ_SCANLINE) == 0 || (d3dcaps.PresentationIntervals & D3DPRESENT_INTERVAL_ONE) == 0) {
			if (config.use_direct3d != 0) config.use_direct3d = 2;
		}
	}
	return hre;
}

void EMU::create_d3dofflinesurface()
{
	HRESULT hre;
	if (pD3Device) {
		hre = pD3Device->CreateOffscreenPlainSurface(screen_width, screen_height, D3DFMT_X8R8G8B8, D3DPOOL_SYSTEMMEM, &pD3Dorigin, NULL);
		if (hre != D3D_OK) {
			release_d3device();
		}
		hre = pD3Device->CreateOffscreenPlainSurface(source_width, source_height, D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT, &pD3Dmixsuf, NULL);
		if (hre != D3D_OK) {
			release_d3device();
		}
#ifdef USE_SCREEN_ROTATE
		hre = pD3Device->CreateOffscreenPlainSurface(screen_height, screen_width, D3DFMT_X8R8G8B8, D3DPOOL_SYSTEMMEM, &pD3Drotate, NULL);
		if (hre != D3D_OK) {
			release_d3device();
		}

		if(config.monitor_type) {
			pD3Dsource = pD3Drotate;
		} else
#endif
		{
			pD3Dsource = pD3Dorigin;
		}
	}
}

HRESULT EMU::reset_d3device(HWND hWnd)
{
	HRESULT hre = D3D_OK;

	hre = pD3Device->Reset(&d3dpp);
	if (hre == D3DERR_DEVICELOST) {
		return hre;
	}
	if (hre != D3D_OK) {
		// re create
		release_d3device();

		hre = create_d3device(hWnd);
		if (hre == D3D_OK) {
			create_d3dofflinesurface();
		}
	}
	return hre;
}

void EMU::release_d3device()
{
#ifdef USE_SCREEN_ROTATE
	if (pD3Drotate != NULL) {
		pD3Drotate->Release();
		pD3Drotate = NULL;
	}
#endif
	pD3Dsource = NULL;

	if (pD3Dorigin != NULL) {
		pD3Dorigin->Release();
		pD3Dorigin = NULL;
		lpD3DBmp = NULL;
	}
	if (pD3Dmixsuf != NULL) {
		pD3Dmixsuf->Release();
		pD3Dmixsuf = NULL;
	}
	if (pD3Device != NULL) {
		pD3Device->Release();
		pD3Device = NULL;
	}
}

void EMU::terminate_d3device()
{
	if (pD3D != NULL) {
		pD3D->Release();
	}
}

void EMU::set_d3dpresent_interval()
{
	if (config.use_direct3d == 2) {
		d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	} else {
		d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
	}
}

#endif // USE_DIRECT3D

inline void EMU::lock_screen()
{
	mux_update_screen->lock();
	screen_changing = true;
}

inline void EMU::unlock_screen()
{
	screen_changing = false;
	mux_update_screen->unlock();
}

