/** @file win32_sound.cpp

	Skelton for retropc emulator

	@author Takeda.Toshiya
	@date   2006.08.18 -

	@brief [ win32 sound ]
*/

#include "emu.h"
#include "vm/vm.h"
#include "video/rec_audio.h"

#undef USE_SOUND_NOTIFY

#define DSOUND_SAMPLE (sound_samples)
#define DSOUND_BUFFER_SIZE (DWORD)(DSOUND_SAMPLE << 3)
#define DSOUND_BUFFER_HALF (DWORD)(DSOUND_SAMPLE << 2)

#define SOUND_RECORD_IN_MIXER

#ifdef USE_AUDIO_U8
#define BITS_PER_SAMPLE	8
#else
#define BITS_PER_SAMPLE	16
#endif

void EMU::EMU_SOUND()
{
	sound_rate = 0;
	sound_samples = 0;
	sound_latency_half = 0;
	sound_ok = sound_started = sound_finished = now_mute = false;
	sound_prev_time = 0;
	sound_nt_event = NULL;
	sound_threadid = 0;
	sound_thread = NULL;
	lpdsnt = NULL;
	lpds = NULL;
	lpdsp = NULL;

	mux_sound_nt = new CMutex();

	rec_audio = new REC_AUDIO(this);
}

void EMU::initialize_sound(int rate, int samples, int latency)
{
	sound_rate = rate;
	sound_samples = samples;
	sound_latency_half = latency / 2;
	vm->initialize_sound(sound_rate, sound_samples);

	// initialize direct sound
	PCMWAVEFORMAT pcmwf;
	DSBUFFERDESC dsbd;
	WAVEFORMATEX wfex;

	if(FAILED(DirectSoundCreate(NULL, &lpds, NULL))) {
		return;
	}
	if(FAILED(lpds->SetCooperativeLevel(main_window_handle, DSSCL_PRIORITY))) {
		return;
	}

	// primary buffer
	ZeroMemory(&dsbd, sizeof(dsbd));
	dsbd.dwSize = sizeof(dsbd);
	dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;
	if(FAILED(lpds->CreateSoundBuffer(&dsbd, &lpdsp, NULL))) {
		return;
	}
	ZeroMemory(&wfex, sizeof(wfex));
	wfex.wFormatTag = WAVE_FORMAT_PCM;
	wfex.nChannels = 2;
	wfex.wBitsPerSample = BITS_PER_SAMPLE;
	wfex.nSamplesPerSec = sound_rate;
	wfex.nBlockAlign = wfex.nChannels * wfex.wBitsPerSample / 8;
	wfex.nAvgBytesPerSec = wfex.nSamplesPerSec * wfex.nBlockAlign;
	if(FAILED(lpdsp->SetFormat(&wfex))) {
		return;
	}

	// secondary buffer
	ZeroMemory(&pcmwf, sizeof(pcmwf));
	pcmwf.wf.wFormatTag = WAVE_FORMAT_PCM;
	pcmwf.wf.nChannels = 2;
	pcmwf.wBitsPerSample = BITS_PER_SAMPLE;
	pcmwf.wf.nSamplesPerSec = sound_rate;
	pcmwf.wf.nBlockAlign = pcmwf.wf.nChannels * pcmwf.wBitsPerSample / 8;
	pcmwf.wf.nAvgBytesPerSec = pcmwf.wf.nSamplesPerSec * pcmwf.wf.nBlockAlign;
	ZeroMemory(&dsbd, sizeof(dsbd));
	dsbd.dwSize = sizeof(dsbd);
	dsbd.dwFlags = DSBCAPS_STICKYFOCUS | DSBCAPS_GETCURRENTPOSITION2;
#ifdef USE_SOUND_NOTIFY
	dsbd.dwFlags |= DSBCAPS_CTRLPOSITIONNOTIFY;
#endif
	dsbd.dwBufferBytes = DSOUND_BUFFER_SIZE;
	dsbd.lpwfxFormat = (LPWAVEFORMATEX)&pcmwf;
	if(FAILED(lpds->CreateSoundBuffer(&dsbd, &lpdsb, NULL))) {
		return;
	}

#ifdef USE_SOUND_NOTIFY
	// create sound notify evnet
	if (FAILED(lpdsb->QueryInterface(IID_IDirectSoundNotify, (VOID**)&lpdsnt))) {
		out_logf(LOG_ERROR,_T("EMU::initialize_sound: QueryInterface %ld"), GetLastError());
		return;
	}
	if ((sound_nt_event = CreateEvent(NULL, FALSE, FALSE, _T("emu_sound_notify"))) == NULL) {
		out_logf(LOG_ERROR,_T("EMU::initialize_sound: CreateEvent %ld"), GetLastError());
		return;
	}

	// set fire point on sound buffer
	DSBPOSITIONNOTIFY dsnt[2];

	dsnt[0].dwOffset = 0;
	dsnt[0].hEventNotify = sound_nt_event;
	dsnt[1].dwOffset = DSOUND_BUFFER_HALF;
	dsnt[1].hEventNotify = sound_nt_event;

	lpdsnt->SetNotificationPositions(2, dsnt);
#endif

	// start new thread
	if ((sound_thread = CreateThread(NULL, 0, sound_proc, (LPVOID)this, 0, &sound_threadid)) == NULL) {
		out_logf(LOG_ERROR,_T("EMU::initialize_sound: CreateThread %ld"), GetLastError());
		return;
	}

	out_logf(LOG_DEBUG, _T("sound ok: rate:%d samples:%d latency:%d"), rate, samples, latency);
	sound_ok = true;
}
void EMU::initialize_sound(int rate, int samples)
{
	initialize_sound(rate, samples, 100);
}

void EMU::initialize_sound()
{
	// load sound config
#ifdef SUPPORT_SOUND_FREQ_55467HZ
	// PC-8801/9801 series
	static int freq_table[8] = {2000, 4000, 8000, 11025, 22050, 44100, 55467, 96000};
#else
	static int freq_table[8] = {2000, 4000, 8000, 11025, 22050, 44100, 48000, 96000};	// Hz
#endif
	static int late_table[5] = {50, 100, 200, 300, 400};	// ms

	if(!(0 <= config.sound_frequency && config.sound_frequency < 8)) {
		config.sound_frequency = 6;	// default: 48KHz
	}
	if(!(0 <= config.sound_latency && config.sound_latency < 5)) {
		config.sound_latency = 1;	// default: 100msec
	}
	int frequency = freq_table[config.sound_frequency];
	int samples = (int)((double)(frequency * late_table[config.sound_latency]) / 1000.0 + 0.5);
	int latency = late_table[config.sound_latency];

	initialize_sound(frequency, samples, latency);
}

void EMU::release_sound()
{
	// release direct sound
	if(lpdsnt) {
		lpdsnt->Release();
	}
	if(lpdsp) {
		lpdsp->Release();
	}
	if(lpdsb) {
		lpdsb->Release();
	}
	if(lpds) {
		lpds->Release();
	}
	lpdsnt = NULL;
	lpdsp = NULL;
	lpdsb = NULL;
	lpds = NULL;

	// stop recording
	stop_rec_sound();

	delete mux_sound_nt;

	delete rec_audio;
}

void EMU::release_sound_on_emu_thread()
{
	// stop recording
	stop_rec_sound();
}

void EMU::start_sound()
{
	if(sound_ok) {
		// start play
		if(!sound_started) {
			lpdsb->Play(0, 0, DSBPLAY_LOOPING);
			sound_started = true;
		}
	}
}

void EMU::end_sound()
{
	if (sound_started) {
		sound_finished = true;
		// stop sound
		lpdsb->Stop();
	}
	if (sound_ok) {
		sound_ok = false;
#ifdef USE_SOUND_NOTIFY
		// close event
		if (sound_nt_event) {
			CloseHandle(sound_nt_event);
		}
#endif

		// wait thread
		if (WaitForSingleObject(sound_thread, 10000) == WAIT_TIMEOUT) {
			// kill thread by force
			if (sound_thread) {
				TerminateThread(sound_thread, 0);
			}
		}
	}
}

//void EMU::update_sound(int* extra_frames)
//{
//}

/// @attention called by another thread
DWORD EMU::update_sound_th()
{
//	out_debug(_T("EMU::update_sound_nt %ld"),timeGetTime());
	DWORD play_c, write_c, offset, size1, size2, next_wait;
	WORD *ptr1, *ptr2;

	// need finish?
	if (sound_finished) {
		return -1;
	}

	// check current playing position
	if(FAILED(lpdsb->GetCurrentPosition(&play_c, &write_c))) {
		out_log(LOG_ERROR,_T("EMU::update_sound_th GetCurrentPosition failed."));
		return -1;
	}
	if(play_c < DSOUND_BUFFER_HALF) {
		offset = DSOUND_BUFFER_HALF;
		next_wait = DSOUND_BUFFER_HALF + (DSOUND_BUFFER_HALF / 4) - play_c;
	} else {
		offset = 0;
		next_wait = DSOUND_BUFFER_SIZE + (DSOUND_BUFFER_HALF / 4) - play_c;
	}

	// sound buffer must be updated
	int extra_frames = 0;

	lock_sound_buffer();
	int16_t* sound_buffer = vm->create_sound(&extra_frames, DSOUND_SAMPLE);
	unlock_sound_buffer();

	if(lpdsb->Lock(offset, DSOUND_BUFFER_HALF, (void **)&ptr1, &size1, (void**)&ptr2, &size2, 0) == DSERR_BUFFERLOST) {
//		out_log(LOG_ERROR,_T("EMU::update_sound DSERR_BUFFERLOST"));
		lpdsb->Restore();
	}
	if(sound_buffer) {
		if(ptr1) {
			CopyMemory(ptr1, sound_buffer, size1);
		}
		if(ptr2) {
			CopyMemory(ptr2, sound_buffer + size1, size2);
		}
	}
	lpdsb->Unlock(ptr1, size1, ptr2, size2);
//	first_half = !first_half;

//	out_logf(LOG_DEBUG,_T("EMU::update_sound off:%ld playc:%ld writec:%ld size1:%ld size2:%ld next:%ld %dms"),offset,play_c,write_c,size1,size2,next_wait,(next_wait * 1000 / sound_rate / 4));

	return (next_wait * 1000 / sound_rate / 4);	// ms
}

DWORD WINAPI EMU::sound_proc(LPVOID lpParameter)
{
	EMU *emu = (EMU *)lpParameter;
	bool working = true;
	DWORD dwResult = 0;

#ifdef USE_SOUND_NOTIFY
	while(working){
		// wait event
		dwResult = MsgWaitForMultipleObjects(1, &emu->sound_nt_event, FALSE, INFINITE, QS_ALLEVENTS);

		switch(dwResult) {
		case WAIT_OBJECT_0 + 0:
			emu->update_sound_th();
			break;
		}
	}
#else
	while(working) {
		CDelay(dwResult);
		if ((dwResult = emu->update_sound_th()) == -1) {
			working = false;
			break;
		}
	}
#endif
	return 0;
}

void EMU::mute_sound(bool mute)
{
	if (mute) {
		now_mute = true;
	} else {
		if(!vm_pause) {
			now_mute = false;
#if 0
			if (sound_started) {
				lpdsb->Play(0, 0, DSBPLAY_LOOPING);
			}
#endif
		}
	}
}

void EMU::set_volume(int volume)
{
//	if (lpdsb) {
//		lpdsb->SetVolume((LONG)volume);
//	}
	// set volume for sound devices.
	if (vm) {
		vm->set_volume();
	}
}

uint32_t EMU::adjust_sound_pos(uint32_t msec)
{
	DWORD play_c, write_c;
	DWORD cur_sec = 0;

	while(cur_sec < (msec + 200)) {
		// check current playing position
		lpdsb->GetCurrentPosition(&play_c, &write_c);
		if ((DSOUND_BUFFER_HALF / 10) < play_c && play_c < (DSOUND_BUFFER_HALF * 4 / 10) && msec <= (cur_sec + 200)) {
			break;
		}
		CDelay(10);
		cur_sec += 10;
	}
	out_debugf(_T("EMU::adjust_sound: p:%ld %ldms"), play_c, cur_sec);
	return cur_sec;
}

void EMU::start_rec_sound(int type)
{
#ifdef USE_REC_AUDIO
	rec_audio->Start(type, sound_rate, false);
#endif
}

void EMU::stop_rec_sound()
{
#ifdef USE_REC_AUDIO
	rec_audio->Stop();
#endif
}

void EMU::restart_rec_sound()
{
#ifdef USE_REC_AUDIO
	rec_audio->Restart();
#endif
}

#if 0
void EMU::record_rec_sound(uint8_t *buffer, int samples)
{
#ifdef USE_REC_AUDIO
	rec_audio->Record(buffer, samples);
#endif
}

void EMU::record_rec_sound(int16_t *buffer, int samples)
{
#ifdef USE_REC_AUDIO
	rec_audio->Record(buffer, samples);
#endif
}
#endif

void EMU::record_rec_sound(int32_t *buffer, int samples)
{
#ifdef SOUND_RECORD_IN_MIXER
#ifdef USE_REC_AUDIO
	rec_audio->Record(buffer, samples);
#endif
#endif
}

bool EMU::now_rec_sound()
{
	return rec_audio->NowRecording();
}

const _TCHAR **EMU::get_rec_sound_codec_list(int type)
{
#ifdef USE_REC_AUDIO
	return rec_audio->GetCodecList(type);
#else
	return NULL;
#endif
}

bool *EMU::get_now_rec_sound_ptr()
{
	return rec_audio->GetNowRecordingPtr();
}

bool EMU::rec_sound_enabled(int type)
{
#ifdef USE_REC_AUDIO
	return rec_audio->IsEnabled(type);
#else
	return false;
#endif
}

void EMU::lock_sound_buffer()
{
	mux_sound_nt->lock();
}
void EMU::unlock_sound_buffer()
{
	mux_sound_nt->unlock();
}
